# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from collections import abc
from typing import Any

from canonical.lib.utils import jsonpath
from .stringtype import StringType


class JSONPath(StringType):

    @classmethod
    def validate(cls, v: str, _: Any = None):
        if not str.startswith(v, '/'):
            raise ValueError("A JSON path must start with a slash.")
        return cls(v)

    def get(self, mapping: abc.Mapping[str, Any]) -> Any:
        try:
            value = jsonpath(mapping, self)
        except (KeyError, TypeError):
            value = None
        return value


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Generic
from typing import TypeVar

from pydantic_core import core_schema
from pydantic_core import CoreSchema
from pydantic import GetJsonSchemaHandler
from pydantic.json_schema import JsonSchemaValue


I = TypeVar('I')
T = TypeVar('T')


class ResourceIdentifier(Generic[I, T]):
    __module__: str = 'canonical'
    openapi_example: Any
    openapi_title: str


    @classmethod
    def __get_pydantic_core_schema__(cls, *_: Any) -> CoreSchema:
        return core_schema.json_or_python_schema(
            json_schema=core_schema.str_schema(max_length=128),
            python_schema=core_schema.union_schema([
                core_schema.is_instance_schema(cls),
                core_schema.chain_schema([
                    core_schema.union_schema([
                        core_schema.str_schema(),
                        core_schema.int_schema(),
                        core_schema.dict_schema()
                    ]),
                    core_schema.no_info_plain_validator_function(cls)
                ]),
            ]),
            serialization=core_schema.plain_serializer_function_ser_schema(str)
        )

    @classmethod
    def __get_pydantic_json_schema__( # pragma: no cover
        cls,
        core_schema: CoreSchema,
        handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue: # pragma: no cover
        json_schema = handler(core_schema)
        json_schema = handler.resolve_ref_schema(json_schema)
        json_schema['examples'] = [cls.openapi_example]
        return json_schema

    def __eq__(self, key: object) -> bool: # pragma: no cover
        return all([
            isinstance(key, type(self)),
            hash(self) == hash(key)
        ])


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Iterable
from typing import Iterator
from typing import ItemsView
from typing import Protocol

from .idatastorekey import IDatastoreKey


class IDatastoreEntity(Protocol):
    __module__: str = 'cbra.ext.google.types'
    key: IDatastoreKey
    def items(self) -> ItemsView[str, Any]: ...
    def update(self, obj: Iterable[tuple[str, Any]] | dict[str, Any]) -> None: ...
    def __iter__(self) -> Iterator[tuple[str, Any]]: ...
    def __len__(self) -> int: ...
    def __getitem__(self, k: str) -> Any: ...
    def __setitem__(self, k: str, v: Any) -> None: ...


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from . import protocols
from .address import Address
from .addressee import Addressee
from .condition import Condition
from .deliverycontact import DeliveryContact
from .deliverypoint import DeliveryPoint
from .displayname import DisplayName
from .domainname import DomainName
from .emailaddress import EmailAddress
from .honorific import HonorificEnum
from .hostname import Hostname
from .httpresourcelocator import HTTPResourceLocator
from .incoterm import Incoterm
from .incoterm import IncotermEnum
from .incoterm import IncotermLiteral
from .iso3166 import ISO3166Alpha2
from .objectmeta import ObjectMeta
from .persistedmodel import PersistedModel
from .personalname import PersonalName
from .phonenumber import Phonenumber
from .pythonsymbol import PythonSymbol
from .resourceidentifier import ResourceIdentifier
from .resourcename import ResourceName
from .resourcename import TypedResourceName
from .resourcestate import CurrentResourceState
from .resourcestate import ResourceState
from .stringtype import StringType
from .text import Text
from .transitioningresource import TransitioningResource
from .unixtimestamp import UnixTimestamp
from .vatrate import VATRate
from .versionedresource import VersionedResource


__all__: list[str] = [
    'protocols',
    'Address',
    'Addressee',
    'CurrentResourceState',
    'Condition',
    'DeliveryContact',
    'DeliveryPoint',
    'DisplayName',
    'DomainName',
    'EmailAddress',
    'HonorificEnum',
    'Hostname',
    'HTTPResourceLocator',
    'Incoterm',
    'IncotermEnum',
    'IncotermLiteral',
    'ISO3166Alpha2',
    'ObjectMeta',
    'PersistedModel',
    'PersonalName',
    'Phonenumber',
    'PythonSymbol',
    'ResourceIdentifier',
    'ResourceName',
    'ResourceState',
    'StringType',
    'Text',
    'TransitioningResource',
    'TypedResourceName',
    'UnixTimestamp',
    'VATRate',
    'VersionedResource',
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic

from canonical.ext.api import APIResourceModel
from .types import Permission
from .types import PermissionWildcard


class BaseRole(APIResourceModel[str], abstract=True):
    permissions: set[PermissionWildcard | Permission] = pydantic.Field(
        default_factory=set,
        description=(
            "The set of permissions granted to this role. May contain "
            "wildcard permissions."
        )
    )

    def is_global(self) -> bool:
        return self.kind == 'ClusterRole'


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import TypeVar

import pydantic

from canonical.lib.types import StringType


M = TypeVar('M', bound=pydantic.BaseModel)


class JWECompactEncoded(StringType):

    @classmethod
    def validate(cls, v: str):
        if not v.count('.') == 4:
            raise ValueError("Invalid JWE Compact Encoding.")
        return cls(v)

    def compact(self):
        return self

    def jose(self, factory: type[M]) -> M:
        protected, key, iv, ct, tag = str.split(self, '.')
        return factory.model_validate({
            'aad': str.encode(protected, 'ascii'),
            'ciphertext': ct,
            'iv': iv,
            'mode': 'compact',
            'protected': protected,
            'recipients': [{'encrypted_key': key}],
            'tag': tag,
        })

    def __repr__(self):
        return f'<{type(self).__name__}>'


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import re
from typing import Any
from typing import TypeVar

from pydantic_core import CoreSchema
from pydantic_core import core_schema
from pydantic.json_schema import JsonSchemaValue
from pydantic import GetJsonSchemaHandler


T = TypeVar('T')


class BytesType(bytes):
    """Base class for bytes types."""
    __module__: str = 'canonical'
    encoding: str = 'ascii'
    lowercase: bool = False
    max_length: int | None = None
    min_length: int | None = None
    openapi_title: str | None = None
    openapi_format: str | None = None
    patterns: re.Pattern[Any] | list[re.Pattern[Any]] = []
    pattern: str | None = None
    strip_whitespace: bool = True

    @classmethod
    def __get_pydantic_core_schema__(cls, *_: Any) -> CoreSchema:
        return core_schema.json_or_python_schema(
            json_schema=cls.__default_schema__(),
            python_schema=core_schema.union_schema([
                core_schema.is_instance_schema(cls),
                core_schema.chain_schema([
                    core_schema.union_schema([
                        core_schema.chain_schema([
                            core_schema.is_instance_schema(str),
                            core_schema.no_info_plain_validator_function(lambda v: v.encode(cls.encoding))
                        ]),
                        core_schema.is_instance_schema(bytes)
                    ]),
                    cls.__default_schema__(),
                    core_schema.no_info_plain_validator_function(cls.transform),
                    core_schema.no_info_plain_validator_function(cls.validate),
                ])
            ]),
            serialization=core_schema.plain_serializer_function_ser_schema(str)
        )

    @classmethod
    def __get_pydantic_json_schema__(
        cls,
        _: CoreSchema,
        handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        return handler(cls.__default_schema__())

    @classmethod
    def __default_schema__(cls):
        return core_schema.bytes_schema(
            max_length=cls.max_length,
            min_length=cls.min_length,
        )

    @classmethod
    def transform(cls, v: str) -> str:
        return v

    @classmethod
    def validate(cls, v: bytes) -> bytes:
        return cls(v)

    def __repr__(self) -> str: # pragma: no cover
        return f'<{type(self).__name__}: {str(self)}>'


# Copyright (C) 2023-2024 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

from .apimodel import APIModel
from .apimodelfield import APIModelField
from .apiresource import APIResource as APIResourceModel
from .apimodelinspector import APIModelInspector
from .apirootresource import APIRootResource as APIRootResourceModel
from .apiversioned import APIVersioned
from .apiversionedmeta import APIVersionedMeta
from .clusterobjectreference import ClusterObjectReference
from .localobjectreference import LocalObjectReference
from .objectmeta import ObjectMeta
from .objectreference import ObjectReference
from .ownerreference import OwnerReference
from .refs import *
from .registry import get_meta
from .resourcespec import ResourceSpec
from .resourcestatus import ResourceStatus
from .uidreference import UIDReference


__all__: list[str] = [
    'get_meta',
    'APIModel',
    'APIModelField',
    'APIModelInspector',
    'APIResourceModel',
    'APIRootResourceModel',
    'APIVersioned',
    'APIVersionedMeta',
    'ClusterObjectReference',
    'DefaultInspector',
    'LocalObjectReference',
    'ObjectMeta',
    'ObjectReference',
    'OwnerReference',
    'ResourceSpec',
    'ResourceStatus',
    'TypedLocalObjectReference',
    'UIDReference',
]

APIResourceType = APIResourceModel[Any] | APIRootResourceModel[Any]
DefaultInspector = APIModelInspector()


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import logging
from typing import Any
from typing import Callable
from typing import Coroutine

import fastapi
import fastapi.params
import fastapi.routing
import pydantic
from fastapi.dependencies.utils import get_parameterless_sub_dependant
from fastapi.exceptions import HTTPException
from starlette.routing import request_response # type: ignore

from canonical.ext.resource import Error
from canonical.utils.http import MediaTypeSelector
from .response import Response


class APIRoute(fastapi.routing.APIRoute):
    logger: logging.Logger = logging.getLogger('uvicorn')
    media_types: MediaTypeSelector = MediaTypeSelector({
        'text/html',
        'application/json',
        'application/yaml'
    })

    def inject(
        self,
        dependency: fastapi.params.Depends,
        index: int = -1
    ) -> None:
        dependant = get_parameterless_sub_dependant(
            depends=dependency,
            path=self.path_format
        )
        self.dependant.dependencies.insert(0, dependant)
        self.app = request_response(self.get_route_handler())

    def get_route_handler(self) -> Callable[[fastapi.Request], Coroutine[Any, Any, fastapi.Response]]:
        handler = super().get_route_handler()

        async def f(request: fastapi.Request) -> fastapi.Response:
            response: Response[Any]
            try:
                response = await handler(request) # type: ignore
            except pydantic.ValidationError as e:
                # TODO: This should normally be a validation error, but some in
                # some implementations, FastAPI does not catch the validation
                # error (see QueryModelMixin).
                errors: list[dict[str, Any]] = list([dict(x) for x in e.errors()])
                field_errors: list[dict[str, Any]] = []
                if not any([x.get('type') == 'extra_forbidden' for x in errors]):
                    raise
                for error in errors:
                    field_errors.append({
                        'message': error['msg'],
                        'schema_path': tuple(error['loc'] or []),
                        'type': error['type'],
                        'data': error['input']
                    })
                assert request.client
                response = Response(
                    media_type=self.media_types.select(request.headers.get('Accept')),
                    status_code=422,
                    content=Error.factory({
                        'detail': "Illegal additional inputs.",
                        'request': {
                            'url': str(request.url),
                            'host': request.client.host,
                        },
                        'fieldErrors': field_errors,
                        'status_code': 422,
                    })
                )
            except fastapi.exceptions.RequestValidationError as e:
                # TODO: Create a decent error handler.
                errors: list[dict[str, Any]] = list(e.errors())
                status_code: int | None = None
                detail: str | None = None
                field_errors: list[dict[str, Any]] = []
                while errors:
                    error = errors.pop()
                    location, *fields = error.get('loc') or [None]
                    if location is None:
                        raise
                    if location in {'body', 'query'} and fields:
                        status_code = 422
                        field_errors.append({
                            'message': error['msg'],
                            'schema_path': tuple(fields),
                            'type': error['type'],
                            'data': error['input']
                        })
                        continue
                    if location == 'path':
                        # Validation errors in the path should simple return a 404,
                        # because the resource does not exist.
                        status_code = 404
                        break
                    if status_code not in {None, 422}:
                        raise
                assert status_code is not None, e
                assert request.client
                response = Response(
                    media_type=self.media_types.select(request.headers.get('Accept')),
                    status_code=status_code,
                    content=Error.factory({
                        'detail': detail or 'No detailed message provided.',
                        'request': {
                            'url': str(request.url),
                            'host': request.client.host,
                        },
                        'fieldErrors': field_errors,
                        'status_code': status_code,
                    })
                )
            except NotImplementedError as e:
                # TODO: Implement a switch to change the loglevel.
                self.logger.exception("Client requested a feature that was not implemented.")
                assert request.client
                response = Response(
                    media_type=self.media_types.select(request.headers.get('Accept')),
                    status_code=422,
                    content=Error.factory({
                        'detail': "The requested feature is not implemented by the server.",
                        'request': {
                            'url': str(request.url),
                            'host': request.client.host,
                        },
                        'status_code': 422,
                    })
                )
            except HTTPException as e:
                assert request.client
                response = Response(
                    media_type=self.media_types.select(request.headers.get('Accept')),
                    status_code=e.status_code,
                    content=Error.factory({
                        'detail': e.detail,
                        'request': {
                            'url': str(request.url),
                            'host': request.client.host,
                        },
                        'status_code': e.status_code,
                    })
                )
            except Exception as e:
                self.logger.exception("Caught fatal exception.")
                assert request.client
                response = Response(
                    media_type=self.media_types.select(request.headers.get('Accept')),
                    status_code=500,
                    content=Error.factory({
                        'detail': (
                            "The server encountered an unexpected condition "
                            "that prevented it from fulfilling the request."
                        ),
                        'request': {
                            'url': str(request.url),
                            'host': request.client.host,
                        },
                        'status_code': 500,
                    })
                )
            if request.method == 'HEAD':
                # Remove the body from HEAD requests.
                response = Response(
                    status_code=response.status_code,
                    headers=response.headers,
                    media_type=response.media_type,
                )
            return response
        return f


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Callable
from typing import Literal

import pydantic

from canonical.lib.types import PythonSymbol
from .dependencyparameter import DependencyParameter


class Dependency(pydantic.BaseModel):
    kind: Literal['instance', 'factory'] = 'factory'
    _dependency: Any = None
    name: str
    ref: PythonSymbol[Any] = pydantic.Field(
        alias='symbol'
    )
    args: list[DependencyParameter] = []
    kwargs: dict[str, DependencyParameter] = {}

    def build(self):
        """Build the dependency and return a callable that
        is injectable.
        """
        if self._dependency is None:
            injectable = None
            match self.kind:
                case 'instance': injectable = self.build_instance()
                case _:
                    raise NotImplementedError
            assert injectable is not None
            self._dependency = injectable
        return self._dependency

    def build_instance(self) -> Callable[..., Any]:
        """Build a singleton that lives throughout the application
        lifecycle.
        """
        args: list[Any] = [arg.resolve() for arg in self.args]
        kwargs: dict[str, Any] = {k: v.resolve() for k, v in self.kwargs.items()}
        instance = self.ref.value(*args, **kwargs)
        return lambda: instance


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import collections
import inspect
from typing import cast
from typing import Any

import fastapi
import fastapi.params
import pydantic


class QueryModelMixin:

    @classmethod
    def as_query(cls) -> Any:
        """Create a :class:`fastapi.params.Depends` instance
        that obtains the models' input fields from query
        parameters.
        """
        cls = cast(type[pydantic.BaseModel], cls)
        def f(request: fastapi.Request, **kwargs: Any):
            return cls.model_validate(request.query_params)

        sig = inspect.signature(f)
        params: dict[str, inspect.Parameter] = collections.OrderedDict({
            'request': sig.parameters['request']
        })
        for attname, field in cls.model_fields.items():
            if attname == 'request':
                raise ValueError(
                    f"{cls.__name__} specifies the field 'request', which is "
                    "a reserved name."
                )
            if not field.annotation or field.exclude:
                continue
            params[attname] = inspect.Parameter(
                kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
                name=attname,
                annotation=field.annotation,
                default=fastapi.Query(
                    default=field.default,
                    default_factory=field.default_factory,
                    alias=field.alias,
                    title=field.title,
                    description=field.description
                )
            )

        setattr(f, '__signature__', sig.replace(parameters=list(params.values())))
        return fastapi.Depends(f)


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Generic
from typing import TypeVar

from pydantic import GetJsonSchemaHandler
from pydantic.json_schema import JsonSchemaValue
from pydantic_core import CoreSchema
from pydantic_core import core_schema

from canonical.lib.utils import import_symbol

T = TypeVar('T')


class PythonSymbol(Generic[T]):
    value: T

    @classmethod
    def __get_pydantic_core_schema__(cls, *_: Any) -> CoreSchema:
        return core_schema.json_or_python_schema(
            json_schema=core_schema.str_schema(),
            python_schema=core_schema.union_schema([
                core_schema.is_instance_schema(cls),
                core_schema.chain_schema([
                    core_schema.str_schema(),
                    core_schema.no_info_plain_validator_function(cls.fromqualname)
                ]),
            ]),
            serialization=core_schema.plain_serializer_function_ser_schema(str)
        )

    @classmethod
    def __get_pydantic_json_schema__(
        cls,
        _: CoreSchema,
        handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        return handler(core_schema.str_schema(max_length=128))

    @classmethod
    def fromqualname(cls, qualname: str):
        try:
            return cls(
                qualname,
                symbol=import_symbol(qualname)
            )
        except ImportError:
            raise ValueError(f'Unknown Python symbol: {repr(qualname)}')

    def __init__(self, qualname: str, symbol: T) -> None:
        self.qualname = qualname
        self.value = symbol

    def __str__(self):
        return self.qualname


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import fastapi

from ..resourceclient import ResourceClient
from .impersonationauth import ImpersonationAuth


def ImpersonatingClient(key:str, base_url: str):
    async def f(auth: ImpersonationAuth, request: fastapi.Request):
        async with ResourceClient(auth=auth, base_url=base_url) as client:
            setattr(request.state, key, client)
            yield client
    return fastapi.Depends(f)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic

from .bases import BaseNamespace
from .hierarchicalnamespacespec import HierarchicalNamespaceSpec
from .refs import ProjectParentReference


class Project(
    BaseNamespace[str],
    group='',
    version='v1',
    plural='projects',
    type='webiam.io/project'
):
    model_config = {'populate_by_name': True}

    spec: HierarchicalNamespaceSpec[ProjectParentReference] = pydantic.Field(
        default_factory=HierarchicalNamespaceSpec,
        description="Defines the behavior of the `Namespace`."
    )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

import pydantic

from .jsonwebkey import JSONWebKey


class JSONWebKeySet(pydantic.BaseModel):
    keys: list[JSONWebKey] = []
    _index: dict[str, JSONWebKey] = {}

    def has(self, *, kid: str | None = None, thumbprint: str | None = None) -> bool:
        """Return a boolean indicating if the given `kid` or
        `thumbprint` is in the JSON Web Key Set (JWKS).
        """
        return any([
            kid is not None and f'kid:{kid}' in self._index,
            thumbprint is not None and f'thumbprint:{thumbprint}' in self._index,
        ])

    def get(self, *, kid: str | None = None, thumbprint: str | None = None) -> JSONWebKey:
        if not (bool(kid) ^ bool(thumbprint)):
            raise TypeError(
                "The 'kid' and 'thumbprint' parameters are "
                "mutually exclusive."
            )
        k = ''
        if kid:
            k = f'kid:{kid}'
        if thumbprint:
            k = f'thumbprint:{thumbprint}'
        return self._index[k]

    def model_post_init(self, _: Any) -> None:
        for key in self.keys:
            self._index[f'thumbprint:{key.thumbprint}'] = key
            if key.kid is not None:
                self._index[f'kid:{key.kid}'] = key

    def verify(
        self,
        signature: bytes,
        payload: bytes,
        *,
        kid: str | None = None,
        alg: str | None = None
    ) -> bool:
        if kid is None:
            raise NotImplementedError
        is_valid = False
        if self.has(kid=kid):
            jwk = self.get(kid=kid)
            is_valid = jwk.verify(signature, payload, alg=alg)
        return is_valid


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

from canonical.ext.resource import Error
from canonical.ext.jsonpatch import JSONPatchType
from .default import Default


class Patch(Default[Any]):
    detail = True
    exists = True
    input_model = JSONPatchType
    method = 'PATCH'
    requires_body = True
    verb = 'patch'

    def get_openapi_responses(self) -> dict[int, Any]:
        return {
            **super().get_openapi_responses(),
            409: {
                'model': Error,
                'description': (
                    f'One or more of the patches can not be applied to the {self.model.__name__}.'
                )
            }
        }


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import copy
import inspect
from typing import cast
from typing import get_type_hints
from typing import Any
from typing import Generic
from typing import Literal
from typing import TypeVar
from typing import TYPE_CHECKING

import pydantic
from pydantic.fields import FieldInfo

from canonical.lib.protocols import IBuilder
from ..apimodelinspector import APIModelInspector
from ..apimodelfieldinfo import APIModelFieldInfo
if TYPE_CHECKING:
    from ..apimodel import APIModel


T = TypeVar('T', bound='APIModel')


class APIModelClassBuilder(IBuilder[T], Generic[T]):
    inspector = APIModelInspector()

    def __init__(
        self,
        model: type[T],
        metadata_class: type[pydantic.BaseModel] | None,
        abstract: bool = False,
        namespaced: bool = False,
        **kwargs: Any
    ) -> None:
        self.abstract = abstract
        self.building = model.__mode__
        self.metadata_class = metadata_class
        self.model = model
        self.namespaced = namespaced
        self.fields = model.model_fields

    def add_metadata_field(self, namespaced: bool) -> None:
        raise NotImplementedError

    def build_input_model(self):
        if self.building == 'domain' and self.inspector.is_concrete(self.model):
            self.model.__create_model__ = self.subclass(
                self.model, 'create', {}, name=f'{self.model.__name__}DTO'
            )
            self.on_representation_added('create', self.model.__create_model__)

    def construct(self) -> None:
        for name, field in dict(self.fields).items():
            self.model.contribute_to_class(self.model, name, field)
        if not getattr(self.model, '__create_model__', None):
            self.build_input_model()

    def contribute_annotations(self):
        """Iterate over the models' fields and find out it any of the fields is
        an APIModel and contribute them to the model.
        """
        # TODO: For now only support APIModel types itself, Union, List etc
        # comes later.
        for attname, field in self.fields.items():
            if not self.is_api_model_field(field):
                continue
            self.contribute_field(attname, field, field.annotation)

    def contribute_field(self, attname: str, field: FieldInfo, annotation: Any):
        annotation.contribute_to_class(self.model, attname, field)

    def get_presentation_bases(self) -> tuple[type[Any],]:
        return pydantic.BaseModel,

    def get_subclass_attrs(
        self,
        model: type[pydantic.BaseModel],
        mode: str,
        annotations: dict[str, Any] | None = None
    ) -> dict[str, Any]:
        return {
            'model_config': {**model.model_config, 'extra': 'ignore'},
            **copy.deepcopy(model.model_fields),
            '__annotations__': {
                **get_type_hints(model),
                **self.inspector.get_field_annotations(model),
                **(annotations or {})
            },
            '__metadata_class__': getattr(model, '__metadata_class__', None),
            '__mode__': mode,
            '__module__': model.__module__
        }

    def rebuild(self):
        assert self.model.model_rebuild(force=True)
        self.fields = self.model.model_fields

    def set_field_annotation(self, name: str, annotation: Any):
        self.fields[name].annotation = annotation
        self.rebuild()

    def set_field_default(self, name: str, default: Any):
        self.fields[name].default = default
        if self.fields[name].default_factory is not None:
            self.fields[name].default_factory = None

    def set_field_literal(self, name: str, value: str | list[str] | None, *values: Any):
        if value is None:
            self.set_field_annotation(name, str)
            return
        if not isinstance(value, list):
            value = [value]
        self.set_field_annotation(name, Literal[*tuple([*value, *values])]) # type: ignore

    def subclass(
        self,
        model: type[pydantic.BaseModel],
        mode: str,
        attrs: dict[str, Any],
        name: str | None = None,
        bases: tuple[type, ...] | None = None
    ) -> type[T]:
        """Subclass the configured model for the given presentation
        `mode`.
        """
        from ..apimodel import APIModel
        new_class = type(name or model.__name__, bases or self.get_presentation_bases(), {
            **attrs,
            **self.get_subclass_attrs(model, mode)
        })
        new_class = cast(type[pydantic.BaseModel], new_class)
        for name, field in set(model.model_fields.items()):
            if isinstance(field.annotation, TypeVar):
                raise TypeError(
                    "Can not create mode-specific implementations from "
                    f"generic classes (model: {model.__name__}, field: {name})."
                )
            if not isinstance(field, APIModelFieldInfo):
                continue
            if not field.is_included(mode):
                new_class.model_fields.pop(name, None)
                continue
            if inspect.isclass(field.annotation)\
            and issubclass(field.annotation, APIModel):
                new_field = new_class.model_fields[name]
                assert new_field.annotation
                models = self.inspector.get_models(new_field.annotation)
                new_field.annotation = getattr(models, mode)
        new_class.model_rebuild(force=True)
        return cast(type[T], new_class)

    def on_representation_added(self, mode: str, model: type[T]):
        pass

    async def build(self) -> T:
        raise NotImplementedError

    # Meta
    def is_api_model_field(self, field: FieldInfo):
        return inspect.isclass(field.annotation) and issubclass(field.annotation, self.model)

    def has_metadata(self) -> bool:
        return 'metadata' in self.fields


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import NotRequired
from typing import TypedDict
from typing import Union

from httpx._client import UseClientDefault
from httpx._types import *


class AsyncRequestParams(TypedDict):
    content: NotRequired[RequestContent | None]
    data: NotRequired[RequestData | None]
    files: NotRequired[RequestFiles | None]
    json: NotRequired[Any]
    params: NotRequired[QueryParamTypes | None]
    headers: NotRequired[HeaderTypes | None]
    cookies: NotRequired[CookieTypes | None]
    auth: NotRequired[Union[AuthTypes, UseClientDefault, None]]
    follow_redirects: NotRequired[Union[bool, UseClientDefault]]
    timeout: NotRequired[Union[TimeoutTypes, UseClientDefault]]
    extensions: NotRequired[RequestExtensions | None]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import datetime
import json
import secrets
import os
from typing import Any

from google.auth.credentials import Credentials
from google.cloud.iam_credentials_v1 import IAMCredentialsAsyncClient

from canonical.ext.jose.protocols import ITokenSigner


class ServiceAccountTokenSigner(ITokenSigner):
    auto_now: bool = False
    auto_jti: bool = False
    ttl: int | None = None

    def __init__(self, credential: Credentials):
        self.client = IAMCredentialsAsyncClient(credentials=credential)
        self.email = None

        # TODO: Some tricks to find out the service account email.
        email = getattr(credential, 'service_account_email', None)
        if email in {None, 'default'} and 'GOOGLE_SERVICE_ACCOUNT_EMAIL' in os.environ:
            email = os.environ['GOOGLE_SERVICE_ACCOUNT_EMAIL']
        if email in {None, 'default'}:
            raise ValueError("Unable to discover service account email.")
        self.email = email

    async def sign_claims(self, claims: Any, ttl: int | None = None) -> str:
        """Use the service account credentials to sign a JSON Web
        Token (JWT).
        """
        now = datetime.datetime.now(datetime.timezone.utc)
        ttl = ttl or self.ttl
        if ttl is not None:
            claims['exp'] = int((now + datetime.timedelta(seconds=ttl)).timestamp())
        if self.auto_jti:
            claims['jti'] = secrets.token_urlsafe(48)
        if self.auto_now:
            claims['iat'] = claims['nbf'] = int(now.timestamp())
        claims.setdefault('iss', self.email)
        response = await self.client.sign_jwt( # type: ignore
            name=f'projects/-/serviceAccounts/{self.email}',
            payload=json.dumps(claims)
        )
        return response.signed_jwt


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import json
import urllib.parse
from typing import Any
from typing import Callable
from typing import Coroutine

import fastapi
import fastapi.routing
import starlette.responses

from canonical.ext.oauth.utils import DefaultMediaTypeSelector


class ClientEndpointRouteHandler(fastapi.routing.APIRoute):

    class request_class(fastapi.Request):
        selector = DefaultMediaTypeSelector
        # TODO: We accept both form data and JSON, but fastapi.Form and
        # pydantic models don't work well together. Thus, for requests
        # with form data, the json() method should also parse the
        # form data. TODO: This will break if FastAPI to decides to parse
        # the data and inject dependencies based on the Content-Type
        # header.

        async def body(self) -> bytes:
            if not hasattr(self, "_body"):
                chunks: list[bytes] = []
                async for chunk in self.stream():
                    chunks.append(chunk)
                self._body = b"".join(chunks)
                mt = self.selector.select(self.headers.get('Content-Type'))
                if mt == "application/x-www-form-urlencoded":
                    # TODO: Check charset
                    self._headers = self._headers.mutablecopy()
                    self._headers['Content-Type'] = "application/json"
                    data = dict(urllib.parse.parse_qsl(bytes.decode(self._body)))
                    self._body = str.encode(json.dumps(data))
            return self._body

    def get_route_handler(self) -> Callable[[fastapi.Request], Coroutine[Any, Any, starlette.responses.Response]]:
        handler = super().get_route_handler()

        async def f(request: fastapi.Request) -> starlette.responses.Response:
            request = self.request_class(request.scope, request.receive)

            return await handler(request)

        return f


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Protocol
from typing import Self


class ITokenSigner(Protocol):
    __module__: str = 'canonical.ext.jose'

    def configure(
        self,
        auto_now: bool = False,
        ttl: int | None = None,
        auto_jti: bool = False
    ) -> Self:
        ...

    async def sign_claims(self, claims: Any, ttl: int | None = None) -> str:
        ...


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import fastapi.params

from .dependency import Dependency
from .provider import require


class Requirement(fastapi.params.Depends):

    def __init__(self, name: str):
        self.name = name
        require(self)

    def inject(self, dep: Dependency) -> None:
        super().__init__(dependency=dep.build(), use_cache=True)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Literal

from canonical.ext.jose.types import KeyOpEnum
from canonical.lib.types import Base64

from .base import BaseJWK


class SymmetricEncryptionKey(
    BaseJWK[
        Literal['oct'],
        Literal['dir'],
        Literal[KeyOpEnum.decrypt, KeyOpEnum.encrypt],
        Literal['enc']
    ]
):
    k: Base64


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import inspect
from collections import OrderedDict
from collections.abc import MutableMapping
from typing import Any
from typing import Callable
from typing import Generator

from ._deephash import deephash
from .deferred import Deferred
from .loader import import_symbol


__all__: list[str] = [
    'deephash',
    'import_symbol',
    'Deferred',
]


class class_property:
    __module__: str = 'oauthx.utils'

    def __init__(self, func: Callable[..., Any]):
        self.func = func

    def __get__(self, instance: Any, cls: Any) -> Any:
        return self.func(cls)


def merge_signatures(signatures: list[inspect.Signature]) -> inspect.Signature:
    """Merge signatures to that FastAPI can inject the dependencies."""
    params: dict[str, inspect.Parameter] = OrderedDict()
    for sig in signatures:
        for param in sig.parameters.values():
            if param.name in {'self', 'cls'}:
                continue
            if param.name.startswith('_'):
                continue
            if param.kind == inspect.Parameter.VAR_POSITIONAL:
                continue
            if param.kind == inspect.Parameter.VAR_KEYWORD:
                continue
            params[param.name] = param

    return signatures[0].replace(
        parameters=list(sorted(params.values(), key=lambda p: (p.kind, p.default != inspect._empty))) # type: ignore
    )


def throw(cls: type[Exception], *args: Any, **kwargs: Any):
    raise cls(*args, **kwargs)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from __future__ import annotations
from typing import cast
from typing import Any
from typing import ClassVar
from typing import Self
from typing import TYPE_CHECKING

import pydantic

from canonical.lib.types import SerializableSet
from .apimodel import APIModel
from .apimodelfield import APIModelField
if TYPE_CHECKING:
    from .resource import Resource


class ResourceMetadata(APIModel):
    """The base class for all metadata types."""
    model_config = {'populate_by_name': True}
    allow_on_create: ClassVar[set[str]]
    InputModel: ClassVar[type[Self]]

    annotations: dict[str, Any] = APIModelField(
        default_factory=dict,
        title="Annotations",
        description=(
            "Annotations is an unstructured key value map stored with "
            "a resource that may be set by external tools to store "
            "and retrieve arbitrary metadata. They are not queryable and "
            "should be preserved when modifying objects."
        ),
        when={'create', 'update', 'store', 'view'}
    )

    labels: dict[str, str | None] = APIModelField(
        default_factory=dict,
        title="Labels",
        description=(
            "Map of string keys and values that can be used to organize and "
            "categorize (scope and select) objects."
        ),
        when={'create', 'update', 'store', 'view'}
    )

    tags: SerializableSet[str] = APIModelField(
        default_factory=set,
        description=(
            "An array of tags that may be used to classify an object "
            "if a label or annotation is not applicable."
        ),
        when={'create', 'update', 'store', 'view'}
    )

    @classmethod
    def is_namespaced(cls) -> bool:
        return False

    @classmethod
    def default(cls):
        return cls()

    @classmethod
    def add_to_model(cls, model: type[Resource[Any]]) -> None:
        model.model_fields['metadata'].default = None
        model.model_fields['metadata'].default_factory = cls.default

    @classmethod
    def build_input_model(cls) -> type[Self]:
        fields = set({*set(ResourceMetadata.model_fields), *set(cls.model_fields)})
        attrs: dict[str, Any] = {}
        annotations: dict[str, type | None] = attrs.setdefault('__annotations__', {})
        for name in fields:
            if name not in cls.allow_on_create:
                continue
            field = attrs[name] = cls.model_fields[name]
            annotations[name] = field.annotation
        return cast(type[Self], type(cls.__name__, (ResourceMetadata,), attrs))

    def attach(self, resource: Resource[Any]):
        self._api_version = resource.api_version
        self._kind = resource.kind

    def get_namespace(self) -> str | None:
        return None

    def in_namespace(self, namespace: str | None) -> bool:
        raise NotImplementedError

    def is_labeled(self, names: list[str]) -> bool:
        return all([
            self.labels.get(name) not in {None, 'null'}
            for name in names
        ])

    def merge(self, metadata: 'ResourceMetadata'):
        """Merge :class:`ObjectMeta` `metadata` into this instance."""
        # Only annotations, labels and tags are merged for now.
        self.annotations = {**self.annotations, **metadata.annotations}
        self.labels = {**self.labels, **metadata.labels}
        self.tags = SerializableSet({*self.tags, *metadata.tags})
        return self

    def update_resource_version(self, obj: Resource[Any]):
        pass

    @staticmethod
    def get_metadata_field():
        return pydantic.Field(
            default=...,
            title='Metadata',
            description=(
                "`ObjectMeta` is metadata that all persisted resources "
                "must have, which includes all objects users must create."
            )
        )


# Copyright (C) 2023-2024 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic

from .apimodel import APIModel
from .bases import BaseReference
from .fields import ReferencedName


__all__: list[str] = [
    'LocalObjectReference'
]


class LocalObjectReference(APIModel, BaseReference):
    """A `LocalObjectReference` instance contains enough
    information to let you locate the referenced object
    inside the same namespace.
    """
    _namespace: str | None = pydantic.PrivateAttr(default=None)
    name: ReferencedName

    def as_name(self) -> str:
        return self.name

    def attach_to_namespace(self, namespace: str):
        self._namespace = namespace
        return self

    def get_namespace(self) -> str | None:
        if self._namespace is None:
            raise TypeError("LocalObjectReference must be attached to a namespace.")
        return self._namespace

    def in_namespace(self, namespace: str) -> bool:
        return True

    def is_cluster(self) -> bool:
        return False

    def is_namespaced(self) -> bool:
        return True

    def with_namespace(self, namespace: str):
        self._namespace = namespace
        return self


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from canonical.ext.oauth.bases import Endpoint
from canonical.ext.oauth.models.responses import TokenResponse
from canonical.ext.oauth.params import TokenEndpointClient
from canonical.ext.oauth.params import TokenRequest


class TokenEndpoint(Endpoint):
    name = 'oauth2.token'
    path = '/token'
    response_model = TokenResponse
    summary = 'Token Endpoint'

    async def post(
        self,
        client: TokenEndpointClient,
        params: TokenRequest
    ):
        print(client)
        print(params.model_dump_json(indent=2))


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from __future__ import annotations
import inspect
from typing import Any
from typing import ClassVar
from typing import Literal
from typing import Self
from typing import TypeVar
from typing import Union
from typing import TYPE_CHECKING

import pydantic
from pydantic.fields import FieldInfo

from .primarykey import PrimaryKey
if TYPE_CHECKING:
    from .objectmeta import ObjectMeta
    from .namespacedobjectmeta import NamespacedObjectMeta
    from .resource import Resource
    from .rootresource import RootResource


T = TypeVar('T', bound=pydantic.BaseModel)


class ResourceMeta(pydantic.BaseModel):
    _create_model: type[pydantic.BaseModel] = pydantic.PrivateAttr()
    input_excluded_fields: ClassVar[set[str]] = {'metadata', 'status'}
    api_group: str
    version: str
    kind: str
    namespaced: bool
    plural: str

    @classmethod
    def fromqualname(cls, model: type[Resource[Any] | RootResource[Any]], qualname: str):
        name, version = str.split(qualname, '/')
        if '.' in name:
            plural, group = name.split('.', 1)
        else:
            group = ''
            plural = name
        return cls(
            api_group=group,
            version=version,
            kind=model.__name__,
            namespaced=model.is_namespaced(),
            plural=plural
        )

    @property
    def api_version(self) -> str:
        return f'{self.api_group}/{self.version}'\
            if self.api_group\
            else self.version

    def contribute_to_class(
        self,
        cls: type[Resource[Any] | RootResource[Any]],
        fields: dict[str, FieldInfo],
        root: tuple[type[Resource[Any]], ...] | None = None
    ):
        cls.__meta__ = self
        cls.group = self.api_group
        cls.plural = self.plural
        base_path = self.api_group
        if base_path:
            base_path += '/'
        base_path = f'{base_path}{self.version}'
        if cls.is_namespaced():
            base_path += '/namespaces/{namespace}'
        cls.base_path = f'{base_path}/{self.plural}'

        # Set defaults and annotations.
        metadata_class = None
        if fields:
            fields['api_version'].annotation = Literal[f'{self.api_version}']
            fields['api_version'].default = self.api_version
            fields['kind'].annotation = Literal[f'{self.kind}']
            fields['kind'].default = self.kind

            # Find the metadata class
            metadata_class: Any = fields['metadata'].annotation
            if not inspect.isclass(metadata_class):
                raise NotImplementedError
            metadata_class.add_to_model(cls)
            assert cls.model_rebuild(force=True)
        cls.InputModel = self.build_input_model(cls, metadata_class, fields, root=root) # type: ignore

        cls.KeyType = PrimaryKey.typed(cls) # type: ignore
        #assert cls.model_rebuild(force=True)
        cls.__meta__ = self

    @classmethod
    def build_input_model(
        cls,
        model: type[T],
        meta: type['ObjectMeta[Any]'] | type['NamespacedObjectMeta[Any]'],
        fields: dict[str, FieldInfo],
        root: tuple[type[Resource[Any]], ...] | None = None
    ) -> Any:
        """Create an input model for the given resource implementation."""
        # The default implementation is to exclude the status for and
        # use a simplified ObjectMeta type.
        if root:
            return type(model.__name__, (pydantic.RootModel, ), { # type: ignore
                'model_config': model.model_config,
                '__annotations__': {'root': Union[*[model.InputModel for model in root]]} # type: ignore
            })

        # For Resources, simply create a subclass
        assert meta is not None
        fields = {
            k: v for k, v in fields.items()
            if k not in cls.input_excluded_fields
        }
        annotations: dict[str, Any] = {
            **{name: field.annotation for name, field in fields.items()},
            'metadata': meta.build_input_model(),
        }
        return type(model.__name__, (pydantic.BaseModel,), { # type: ignore
            'model_config': {
                **model.model_config,
                'title': f'Create{model.__name__}Request'
            },
            'metadata': meta.get_metadata_field(),
            '__annotations__': annotations,
            **fields
        })

    def update(self, resource: Any, metadata: Self | None, mode: Literal['replace', 'update']):
        pass


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import TypeVar

from canonical.ext.api import APIResourceType
from .create import Create
from ..params import ResourceRepository


T = TypeVar('T', bound=APIResourceType)


class DefaultCreate(Create[T]):
    creates = True
    detail = False
    exists = False
    method = 'POST'
    requires_body = True
    status_code = 201
    verb = 'create'

    async def handle(
        self,
        repo: ResourceRepository,
        obj: APIResourceType
    ) -> APIResourceType:
        obj = await repo.persist(
            model=self.model,
            new=self.model.model_validate({
                'metadata': {
                    **obj.metadata.model_dump(),
                    'generation': 1,
                    'uid': await repo.allocate(self.model)
                },
                **obj.model_dump(exclude={'metadata'})
            })
        )
        self.logger.debug(
            "Persisted object (kind: %s, uid: %s).",
            obj.kind, obj.metadata.uid
        )
        return obj


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Generic
from typing import TypeVar

import pydantic

from canonical.ext.api import APIResourceModel


T = TypeVar('T')


class BaseClient(APIResourceModel[str], Generic[T], abstract=True):
    spec: T = pydantic.Field(
        default=...,
        description=(
            "Specification of the OAuth 2.x/OpenID Connect client."
        )
    )


# Copyright (C) 2023-2024 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import cast
from typing import Any
from typing import Generic
from typing import TypeVar

import pydantic

from ..fields import APIVersion
from ..fields import Kind


T = TypeVar('T')


class BaseResourceDTO(pydantic.BaseModel, Generic[T]):
    api_version: APIVersion
    kind: Kind

    @property
    def key(self) -> Any:
        raise NotImplementedError

    @property
    def model(self) -> type[T]:
        raise NotImplementedError

    def factory(self, uid: int) -> T:
        # TODO
        model = cast(type[pydantic.BaseModel], self.model)
        data = self.model_dump()
        data['metadata']['generation'] = 1
        data['metadata']['uid'] = uid
        return cast(T, model.model_validate(data))


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import AsyncIterator
from typing import Protocol
from typing import Self
from typing import TypeVar


T = TypeVar('T', covariant=True)


class ITransaction(Protocol):
    __module__: str = 'canonical.protocols'

    async def transaction(
        self,
        transaction: Self | None = None
    ) -> AsyncIterator[Self]:
        ...


# Copyright (C) 2023-2024 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from __future__ import annotations
from typing import Annotated
from typing import Any
from typing import TYPE_CHECKING
from typing import Union
from typing_extensions import Doc

if TYPE_CHECKING:
    from canonical.ext.api import *


APIResourceType = Annotated[
    Union['APIResourceModel[Any]', 'APIRootResourceModel[Any]'],
    Doc("Versioned resource type.")
]

Reference = Annotated[
    Union['LocalObjectReference', 'ObjectReference', 'OwnerReference', 'UIDReference'],
    Doc("Reference type to an object.")
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Annotated

import fastapi

from canonical.ext.api import APIModelInspector
from canonical.ext.crypto.bases import BaseDEK
from canonical.lib.protocols import ICache
from ..utils import request_state as s
from ..utils import inject_state as d


__all__: list[str] = [
    'AccessPolicyServerURL',
    'DefaultCache',
    'DefaultInspector',
]


AccessPolicyServerURL = Annotated[str, d('core_url')]
DataEncryption = Annotated[BaseDEK, s('dek', BaseDEK)]
DefaultCache = Annotated[ICache, s('cache', ICache, True)]
DefaultInspector = Annotated[APIModelInspector, fastapi.Depends(APIModelInspector)]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic

from canonical.ext.api import ResourceStatus


class AuthorizationRequestStateStatus(ResourceStatus):
    code: str | None = pydantic.Field(
        default=None,
        alias='authorizationCode',
        title="Authorization Code",
        description=(
            "The authorization code that may be used to obtain an "
            "access token from the token endpoint."
        )
    )
    access_token: str | None = pydantic.Field(
        default=None,
        alias='accessToken',
        title="Access Token",
        description=(
            "For proxying and obtaining states, the access token received "
            "from the authorization server.\n\nFor issuing states, the "
            "issued access token."
        )
    )

    refresh_token: str | None = pydantic.Field(
        default=None,
        title="Refresh Token",
        alias='refreshToken',
        description=(
            "For proxying and obtaining states, the refresh token received "
            "from the authorization server.\n\nFor issuing states, the "
            "issued refresh token."
        )
    )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Literal
from typing import TypeVar

from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.primitives.asymmetric.padding import AsymmetricPadding
from cryptography.hazmat.primitives.asymmetric.padding import PKCS1v15
from cryptography.hazmat.primitives.hashes import HashAlgorithm
from cryptography.hazmat.primitives.hashes import SHA256
from cryptography.hazmat.primitives.hashes import SHA384
from cryptography.hazmat.primitives.hashes import SHA512

from canonical.ext.jose.types import KeyOpEnum
from .rsakey import RSAKey


A = TypeVar('A')
O = TypeVar('O')
U = TypeVar('U')


class RSAVerificationKey(
    RSAKey[
        Literal[
            'RS256',
            'RS384',
            'RS512',
            'PS256',
            'PS384',
            'PS512',
        ],
        Literal[KeyOpEnum.verify],
        Literal['sig']
    ]
):

    def get_digest_algorithm(self) -> type[HashAlgorithm]:
        match self.alg:
            case 'PS256': alg = SHA256
            case 'PS384': alg = SHA384
            case 'PS512': alg = SHA512
            case 'RS256': alg = SHA256
            case 'RS384': alg = SHA384
            case 'RS512': alg = SHA512
            case _: raise NotImplementedError
        return alg

    def get_padding_class(self) -> type[AsymmetricPadding]:
        match self.alg:
            case 'RS256': padding = PKCS1v15
            case 'RS384': padding = PKCS1v15
            case 'RS512': padding = PKCS1v15
            case _: raise NotImplementedError
        return padding

    def get_padding_params(
        self,
        padding_class: type[AsymmetricPadding],
        digest_class: type[HashAlgorithm],
    ) -> dict[str, Any]:
        if padding_class != PKCS1v15:
            raise NotImplementedError
        return {}

    def verify(
        self,
        signature: bytes,
        payload: bytes,
        *,
        alg: str | None = None
    ) -> bool:
        public = self._public_key
        hasher = self.get_digest_algorithm()()
        try:
            public.verify(
                signature,
                payload,
                padding=self.get_padding(),
                algorithm=hasher
            )
            return True
        except InvalidSignature:
            return False


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Protocol

from canonical.ext.oauth.types import AuthenticationMethod


class ITokenRequest(Protocol):

    def identify(self, client_id: str) -> None:
        ...

    def set_client_secret(self, mode: AuthenticationMethod, secret: str) -> None:
        ...


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import inspect
from typing import cast
from typing import Any
from typing import Generic
from typing import TypeVar

from canonical.lib.protocols import IBuilder
from canonical.utils import class_property
from canonical.utils import merge_signatures


P = TypeVar('P')
T = TypeVar('T', covariant=True)


class ParameterizedBuilder(IBuilder[T], Generic[T, P]):
    __module__: str = 'canonical.ext.builder'
    _params: dict[str, Any]

    @property
    def params(self) -> P:
        return cast(P, self._params)

    @class_property
    def __signature__(cls) -> inspect.Signature:
        return merge_signatures([
            inspect.signature(cls.__init__),
            inspect.signature(getattr(cls, 'setup', lambda: None))
        ])

    def __init__(self):
        self._params = {}

    def set_param(self, name: str, value: Any):
        self._params[name] = value

    async def build(self) -> T:
        raise NotImplementedError


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from __future__ import annotations
import inspect
from typing import cast
from typing import Any
from typing import Callable
from typing import Generic
from typing import TypeVar
from typing import Unpack

import fastapi
import fastapi.params

from canonical.exceptions import ProgrammingError
from canonical.ext.cache import NullCache
from canonical.ext.api import APIResourceType
from canonical.ext.api import DefaultInspector
from canonical.ext.iam import PermissionQuery
from canonical.ext.iam import PermissionSet
from canonical.ext.iam.protocols import IResourceAuthorizationContext
from canonical.ext.resource import Error
from canonical.lib.protocols import ICache
from canonical.lib.protocols import ICommand
from canonical.utils import merge_signatures
from .apiroute import APIRoute
from .params import AcceptedContentType
from .params import RequestAuthorizationContext
from .response import Response
from . import verbs


C = TypeVar('C')
R = TypeVar('R', bound=APIResourceType)
ResponseModel = TypeVar('ResponseModel')


class ResourceRouterMetaclass(type):

    def __getitem__(cls, model: type[R]) -> type['ResourceRouter[R]']:
        return type(f'{model.__name__}', (cls,), { # type: ignore
            'model': model
        })


class ResourceRouter(fastapi.APIRouter, Generic[R]):
    inspector = DefaultInspector
    detail_verbs: set[str] = {
        'authorize',
        'destroy',
        'exists',
        'get',
        'inspect',
        'replace',
        'update',
    }
    response_class: type[Response[R]]
    resource_model: type[R]

    @classmethod
    def new(cls, model: type[R], **kwargs: Any) -> 'ResourceRouter[R]':
        return cls.typed(model)(**kwargs)

    @classmethod
    def typed(cls, model: type[R]) -> type['ResourceRouter[R]']:
        impl: type[ResourceRouter[R]]  = type(f'{model.__name__}Router', (cls,), {
            'resource_model': model
        })
        return impl

    def __init__(
        self,
        enabled_verbs: set[str] | None = None,
        granter: bool = False,
        **kwargs: Any
    ):
        self.commands_enabled: list[tuple[type[ICommand], str, Callable[..., Any]]] = []
        self.meta = self.inspector.inspect(self.resource_model)
        self.verbs_enabled = set(enabled_verbs or [])
        deps: list[Any] = kwargs.setdefault('dependencies', [])
        deps.extend([
            *self.inject_defaults()
        ])
        self.collect_dependencies(deps, None)
        model = self.resource_model

        super().__init__(route_class=APIRoute, **kwargs)
        self.response_class = Response.typed(model)
        self.resource_model = model
        self.verbs: list[verbs.Verb[Any]] = []

        self.verbs.append(verbs.Exists(self.resource_model))
        if 'create' in self.verbs_enabled:
            self.verbs.append(verbs.DefaultCreate(self.resource_model))

        meta = model.__meta__
        if granter:
            self.add_resource_route(
                method='POST',
                verb='authorize',
                detail=True,
                path=meta.get_url(True, ':permissions'),
                endpoint=self.permissions,
                summary=f'Test {model.__name__} permissions',
                description=(
                    f"Get the permissions granted to the authenticated "
                    f"subject on a specific `{model.__name__}` object."
                ),
                response_model=PermissionSet,
                response_description=f"PermissionSet object.",
                responses={
                    401: {
                        'model': Error,
                        'description': (
                            "The provided request credential is expired, not effective"
                            ", or otherwise malformed."
                        )
                    },
                    403: {
                        'model': Error,
                        'description': "Untrusted credential."
                    },
                },
                authenticated=False,
                with_resource=False
            )
        return

    def add_to_application(
        self,
        app: fastapi.APIRouter,
        register: Callable[..., None],
        tags: list[Any],
        authorization: type[IResourceAuthorizationContext] | None = None,
        cache: ICache = NullCache(),
    ):
        for verb in self.verbs:
            verb.add_to_router(self, register, authorization=authorization, cache=cache)
        tags.append({
            'name': self.resource_model.__name__,
            'description': getattr(self.resource_model, '__doc__', None)
        })

        if self.commands_enabled:
            commands = verbs.Command(self.resource_model)
            for command_class, verb, handler in self.commands_enabled:
                commands.add(command_class, verb, handler)
            commands.add_to_router(self, register, authorization=authorization)
        app.include_router(router=self)

    def add_resource_route(
        self,
        *,
        verb: str,
        method: str,
        path: str,
        summary: str,
        description: str,
        endpoint: Any,
        response_model: type[Any] | None = None,
        response_description: str,
        responses: dict[int, Any] | None = None,
        dependencies: list[Any] | None = None,
        detail: bool = False,
        authenticated: bool = True,
        with_resource: bool = True
    ) -> None:
        model = self.resource_model
        responses = dict(responses or {})
        dependencies = dependencies or []
        if verb in self.detail_verbs:
            detail = True
        if verb in {'create', 'replace', 'update', 'authorize'}:
            dependencies.append(AcceptedContentType({'application/yaml', 'application/json'}))
            responses.update({
                415: {
                    'model': Error,
                    'description': "Invalid content type for request body."
                },
            })
        if detail:
            responses.update({
                404: {
                    'model': Error,
                    'description': (
                        f'The {model.__name__} specified by the path parameter(s) '
                        'does not exist.'
                    )
                }
            })
            #if with_resource:
            #    dependencies.insert(0, self.inject_resource())
        if not authenticated:
            responses.pop(401, None)
        self.add_api_route(
            methods=[method],
            path=path,
            endpoint=endpoint,
            summary=summary,
            description=description,
            responses={**responses},
            response_model=self.resource_model or response_model,
            response_model_by_alias=True,
            response_description=response_description,
            dependencies=[
                fastapi.Depends(self.set_verb(verb)),
                *dependencies,
            ],
            tags=[model.__name__],
        )
        #self.register_resource(path, self.resource_model)

    def add_verb(
        self,
        cls: type[verbs.Verb[Any]],
        authenticated: bool = True,
        ttl: int | None = None,
        **kwargs: Unpack[verbs.VerbParameters]
    ) -> None:
        self.verb_factory(
            cls,
            authenticated=authenticated,
            ttl=ttl,
            **kwargs
        )

    def collect_dependencies(self, dependencies: list[Any], verb: str | None):
        pass

    def command(self, model: type[Any], verb: str) -> Callable[..., Any]:
        model = cast(type[ICommand], model)

        def f(func: Callable[..., Any]) -> Any:
            sig = inspect.signature(func)
            args = list(sig.parameters.values())
            fn = f'{func.__name__}()'
            if len(args) < 2:
                raise TypeError(
                    f"Handler function {fn} must accept "
                    "at least two positional parameters."
                )
            if args[0].default != inspect.Parameter.empty:
                raise TypeError(
                    f'First positional argument to {fn} must not '
                    'have a default.'
                )
            if args[0].annotation != self.resource_model:
                raise TypeError(
                    f'First positional argument to {fn} must be '
                    f'annotated with {self.resource_model.__name__}.'
                )
            if args[1].default != inspect.Parameter.empty:
                raise TypeError(
                    f'Second positional argument to {fn} must not '
                    'have a default.'
                )
            if args[1].annotation != model:
                raise TypeError(
                    f'Second positional argument to {fn} must be '
                    f'annotated with {model.__name__}.'
                )
            if args[1].name != 'command':
                raise TypeError(
                    f'Second positional argument to {fn} must be '
                    f'named \'command\'.'
                )

            h = (model, verb, func)
            if h in self.commands_enabled:
                raise ValueError(
                    f"Handler {fn}() is already registered for command "
                    f"{model.__name__} with verb '{verb}'."
                )
            self.commands_enabled.append(h)
            return func
        return f

    def get_media_type(self, request: fastapi.Request) -> str | None:
        return getattr(request.state, 'media_type', None)

    def inject_defaults(self) -> list[fastapi.params.Depends]:
        return [
            fastapi.params.Depends(self.inject_model),
            fastapi.params.Depends(self.inject_namespace),
        ]

    def inject_model(self, request: fastapi.Request):
        setattr(request.state, 'model', self.resource_model)

    def inject_namespace(self, request: fastapi.Request):
        if 'name' in request.path_params:
            setattr(request.state, 'name', request.path_params['name'])
        if 'namespace' in request.path_params:
            setattr(request.state, 'namespace', request.path_params['namespace'])

    def writer_factory(self, model: type[R], func: Any, partial: bool = False):
        async def f(resource: model, *args: Any, **kwargs: Any) -> Any:
            return await func(resource, *args, **kwargs)
        f.__signature__ = merge_signatures([ # type: ignore
            inspect.signature(func),
            inspect.signature(f)
        ])
        return f

    def set_verb(self, verb: str):
        return self.inject_verb(verb)

    def inject_verb(self, verb: str):
        def f(request: fastapi.Request):
            setattr(request.state, 'verb', verb)
        return f

    def subresource(
        self,
        methods: list[str],
        path: str,
        verb: str,
        **kwargs: Any
    ):
        assert not str.startswith(path, '/')
        path = f'/{self.meta.base_path}/{{name}}/{path}'
        dependencies: list[Any] = kwargs.setdefault('dependencies', [])
        dependencies.insert(0, fastapi.Depends(self.inject_verb(verb)))
        return self.api_route(
            path=path,
            methods=methods,
            tags=[self.resource_model.__name__],
            **kwargs
        )
        
    async def permissions(
        self,
        request: fastapi.Request,
        ctx: RequestAuthorizationContext,
        query: PermissionQuery
    ) -> Response[Any]:
        return self.render_to_response(
            request=request,
            status_code=200,
            instance=PermissionSet.model_validate({
                'granted': await ctx.get_permissions({str(p) for p in query.spec.permissions})
            })
        )

    def render_to_response(
        self,
        *,
        request: fastapi.Request,
        instance: Any,
        status_code: int = 200
    ) -> Response[Any]:
        return Response(
            status_code=status_code,
            media_type=self.get_media_type(request),
            content=instance
        )

    def verb(
        self,
        cls: type[verbs.Verb[Any]],
        *,
        verb: str | None = None,
        authenticated: bool = True,
        ttl: int | None = None,
        **kwargs: Unpack[verbs.VerbParameters]
    ) -> Callable[..., Any]:
        def f(func: Callable[..., Any]) -> Callable[..., Any]:
            sig = inspect.signature(func)
            params = list(sig.parameters.values())
            if not params or params[0].name != 'verb':
                raise ProgrammingError(
                    f"The first argument to {func.__module__}.{func.__name__}() "
                    "must be named 'verb'."
                )

            impl: verbs.Verb[[], Any] = type(cls.__name__, (cls,), { # type: ignore
                'handle': func,
                'verb': verb or getattr(cls, 'verb')
            })
            self.verb_factory(impl, authenticated=authenticated, ttl=ttl, **kwargs) # type: ignore
            return func
        return f

    def verb_factory(self, cls: type[verbs.Verb[Any]], *args: Any, **kwargs: Any):
        if any([v.verb == cls.verb for v in self.verbs]):
            raise ValueError(f"Verb '{cls.verb}' is already declared.")
        self.verbs.append(cls(model=self.resource_model, *args, **kwargs)) # type: ignore
        self.verbs_enabled.add(cls.verb)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic

from canonical.lib.types import Base64
from .baseencryptionresult import BaseEncryptionResult


class AESEncryptionResult(BaseEncryptionResult):
    """The result of an encryption operation using the AES algorithm."""
    aad: Base64 = pydantic.Field(
        default=b'',
        title="Additional Authenticated Data (AAD)",
        description=(
            "The Base64-encoded input data to the authenticated encryption "
            "function that is authenticated but not encrypted."
        )
    )

    ct: Base64 = pydantic.Field(
        default=...,
        title="Ciphertext",
        description=(
            "The Base64-encoded encryption result."
        )
    )

    iv: Base64 = pydantic.Field(
        default=...,
        title="Initialization Vector (IV)",
        description=(
            "A binary vector used as the input to initialize the algorithm for "
            "the encryption of a plaintext block sequence to increase security "
            "by introducing additional cryptographic variance and to synchronize "
            "cryptographic equipment. The initialization vector need not be secret."
        )
    )

    tag: Base64 = pydantic.Field(
        default=...,
        title="Authentication tag",
        description=(
            "A cryptographic checksum on data that is designed to reveal both "
            "accidental errors and the intentional modification of the data."
        )
    )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Iterable
from typing import Protocol

from canonical.protocols import IDeferrable
from ..types import PermissionSet


class IResourceAuthorizationContext(IDeferrable, Protocol):

    def has_permissions(self, permissions: Iterable[str]) -> set[str]: ...
    def is_authenticated(self) -> bool: ...
    def is_authorized(self) -> bool: ...
    async def has(self, permissions: str | set[str]) -> bool: ...
    async def get_permissions(self, permissions: set[str]) -> PermissionSet: ...


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import hashlib
import json
import logging
from typing import Any
from typing import Literal
from typing import TypeVar

import pydantic

from canonical.ext.crypto import EncryptionResult
from canonical.ext.crypto import IDataEncryption
from canonical.ext.crypto import NullDataEncryption
from canonical.lib.protocols import ICache


T = TypeVar('T')


class BaseCache(ICache):
    __module__: str = 'canonical.ext.cache'
    dek: IDataEncryption[EncryptionResult]
    logger: logging.Logger = logging.getLogger('uvicorn')

    def __init__(
        self,
        keyalg: Literal['sha256'] | None = None,
        prefix: str | None = None,
        dek: IDataEncryption[EncryptionResult] = NullDataEncryption()
    ):
        self.dek = dek
        self.keyalg = keyalg
        self.prefix = prefix

    def hash(self, value: str, algorithm: str | None = None) -> str:
        if self.prefix:
            value = f'{self.prefix}:{value}'
        algorithm = algorithm or self.keyalg
        if algorithm is not None:
            h = hashlib.new(algorithm)
            h.update(str.encode(value))
            value = h.hexdigest()
        return value

    async def serialize(
        self,
        value: T | EncryptionResult,
        encoder: type[T],
        encrypt: bool = False
    ) -> bytes:
        adapter = pydantic.TypeAdapter[T | EncryptionResult](encoder)
        if encrypt:
            value = await self.dek.encrypt(value, encoder)
        return adapter.dump_json(value)

    async def deserialize(self, value: bytes , decoder: type[T], mode: str) -> T:
        adapter = pydantic.TypeAdapter[T | EncryptionResult](decoder)
        match mode:
            case 'json':
                result = adapter.validate_python(json.loads(value))
            case 'python':
                result = adapter.validate_python(value)
            case _:
                raise NotImplementedError(mode)
        if isinstance(result, EncryptionResult):
            result = await self.dek.decrypt(result, decoder)
        return result

    async def get(
        self,
        key: str,
        decoder: type[T] = bytes,
        validation: Literal['strict', 'ignore'] = 'strict',
        mode: Literal['json'] = 'json',
        keyalg: Literal['sha256'] | None = None
    ) -> T | None:
        key = self.hash(key, algorithm=keyalg)
        value = await self.fetch(key)
        if value is None:
            self.logger.debug(
                self.logger.debug(
                    "Cache miss (key: %s, using: %s)",
                    key, type(self).__name__
                )
            )
            return None
        try:
            self.logger.debug("Cache hit (key: %s, using: %s)", key, type(self).__name__)
            assert isinstance(value, bytes)
            return await self.deserialize(value, decoder, mode)
        except (pydantic.ValidationError, ValueError, TypeError):
            if validation == 'strict':
                raise
            await self.clear(key)
            return None

    async def set(
        self,
        key: str,
        value: Any,
        encoder: type[T] = bytes,
        encrypt: bool = False,
        ttl: int | None = None,
        keyalg: Literal['sha256'] | None = None
    ) -> None:
        self.logger.debug("Cache write (key: %s, using: %s)", key, type(self).__name__)
        value = await self.serialize(value, encoder, encrypt=encrypt)
        await self.put(
            key=self.hash(key, algorithm=keyalg),
            value=value,
            encrypt=encrypt,
            ttl=ttl
        )

    async def clear(self, key: str) -> None:
        raise NotImplementedError

    async def fetch(self, key: str) -> bytes | None:
        raise NotImplementedError

    async def put(
        self,
        key: str,
        value: Any,
        encrypt: bool = False,
        ttl: int | None = None
    ) -> None:
        raise NotImplementedError


# Copyright (C) 2023-2024 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from __future__ import annotations
from typing import Any
from typing import cast
from typing import overload
from typing import Generic
from typing import TypeVar

import pydantic


T = TypeVar('T')


class RootModelDescriptor(Generic[T]):

    def __init__(self, attname: str) -> None:
        self.attname = attname

    @overload
    def __get__(self, obj: None, cls: None) -> RootModelDescriptor[T]:
        ...

    @overload
    def __get__(self, obj: pydantic.RootModel[Any], cls: type[pydantic.RootModel[Any]]) -> T:
        ...

    def __get__(
        self,
        obj: pydantic.RootModel[Any] | None,
        cls: type[pydantic.RootModel[Any]] | None = None
    ) -> RootModelDescriptor[T] | T:
        if obj is None:
            return self
        return cast(T, obj.root.__dict__[self.attname])


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import os
from typing import Annotated
from typing import TypeAlias

import fastapi


__all__: list[str] = ['DatastoreProject']




async def get(request: fastapi.Request) -> str:
    return getattr(
        request.state,
        'google_datastore_project',
        os.environ['GOOGLE_DATASTORE_PROJECT']
    )


DatastoreProject: TypeAlias = Annotated[str, fastapi.Depends(get)]


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic
import pytest

from canonical import DomainName


@pytest.mark.parametrize('v', [
    'com',
    'example.com',
    'www.example.com',
    '_.example.com',
    'localhost'
])
def test_validate(v: str):
    DomainName.validate(v)


@pytest.mark.parametrize('v', [
    'abc@example.com',
    '-foo',
    'foo-',
    '-foo.bar',
    'foo.bar-',
    f'{"a" * 64}',
    f'{"a" * 64}.com',
])
def test_invalid(v: str):
    with pytest.raises(ValueError):
        DomainName.validate(v)


@pytest.mark.parametrize('v', [
    'com',
    'example.com',
    'www.example.com',
    'localhost'
])
def test_as_model_attribute(v: str):
    class Model(pydantic.BaseModel):
        domain: DomainName

    obj = Model.model_validate({'domain': v})
    assert isinstance(obj.domain, DomainName)

@pytest.mark.parametrize('v', [
    '',
    'abc@example.com',
    '-foo',
    'foo-',
    '-foo.bar',
    'foo.bar-',
    f'{"a" * 64}',
    f'{"a" * 64}.com',
    'a' * 254,
])
def test_invalid_model(v: str):
    class Model(pydantic.BaseModel):
        domain: DomainName

    with pytest.raises(pydantic.ValidationError):
        Model.model_validate({'domain': v})


def test_index():
    name = DomainName('foo.bar.baz')
    assert name[0] == 'foo'
    assert name[1] == 'bar'
    assert name[2] == 'baz'


def test_index_slice():
    name = DomainName('foo.bar.baz')
    assert name[1:] == 'bar.baz'
    assert name[:-1] == 'foo.bar'


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .base64 import Base64
from .base64int import Base64Int
from .base64json import Base64JSON
from .bytestype import BytesType
from .domainname import DomainName
from .emailaddress import EmailAddress
from .httprequestref import HTTPRequestRef
from .httpresourcelocator import HTTPResourceLocator
from .jsonpath import JSONPath
from .pythonsymbol import PythonSymbol
from .resourcename import ResourceName
from .resourcename import TypedResourceName
from .serializableset import SerializableSet
from .stringtype import StringType
from .unixtimestamp import UnixTimestamp


__all__: list[str] = [
    'Base64',
    'Base64Int',
    'Base64JSON',
    'BytesType',
    'DomainName',
    'EmailAddress',
    'HTTPRequestRef',
    'HTTPResourceLocator',
    'JSONPath',
    'PythonSymbol',
    'ResourceName',
    'SerializableSet',
    'StringType',
    'TypedResourceName',
    'UnixTimestamp',
]


import contextlib
from typing import Any
from typing import TypeVar

import pydantic
from canonical.protocols import IRepository


T = TypeVar('T', bound='PersistedModel')


class PersistedModel(pydantic.BaseModel):
    _repo: IRepository[Any] = pydantic.PrivateAttr()

    def attach(self: T, storage: IRepository[T]):
        self._repo = storage
        return self

    @contextlib.asynccontextmanager
    async def atomic(self):
        try:
            yield
            await self.persist()
        except Exception:
            raise

    @contextlib.asynccontextmanager
    async def consume(self):
        try:
            yield
            await self.delete()
        except Exception:
            raise

    async def delete(self) -> None:
        await self._repo.delete(self)

    async def persist(self) -> None:
        await self._repo.persist(self)


# Copyright (C) 2023-2024 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import cast
from typing import Any
from typing import Literal

from pydantic.fields import FieldInfo


class APIModelFieldInfo(FieldInfo):
    __slots__ = (*FieldInfo.__slots__, 'when', 'encrypt')

    @classmethod
    def from_annotation(cls, annotation: type[Any]) -> FieldInfo:
        raise NotImplementedError

    @classmethod
    def from_annotated_attribute(cls, annotation: type[Any], default: Any) -> FieldInfo:
        raise NotImplementedError

    @classmethod
    def from_field(cls, *args: Any, **kwargs: Any): # type: ignore
        raise NotImplementedError

    @classmethod
    def from_field_info(cls, field: FieldInfo, **kwargs: Any):
        kwargs.update({
            k: getattr(field, k) for k in field.__slots__
            if not str.startswith(k, '_')
        })
        return cls(**kwargs)

    @classmethod
    def merge_field_infos(cls, *field_infos: FieldInfo, **overrides: Any): # type: ignore
        """Merge `FieldInfo` instances keeping only explicitly set attributes.

        Later `FieldInfo` instances override earlier ones.

        Returns:
            FieldInfo: A merged FieldInfo instance.
        """
        flattened_field_infos: list[FieldInfo] = []
        for field_info in field_infos:
            flattened_field_infos.extend(x for x in field_info.metadata if isinstance(x, FieldInfo))
            flattened_field_infos.append(field_info)
        field_infos = tuple(flattened_field_infos)
        new_kwargs: dict[str, Any] = {}
        metadata = {}
        for field_info in field_infos:
            new_kwargs.update(field_info._attributes_set)
            for x in field_info.metadata:
                if not isinstance(x, FieldInfo):
                    metadata[type(x)] = x
        new_kwargs.update(overrides)
        field_info = cls(**new_kwargs)
        field_info.metadata = list(cast(Any, metadata.values()))
        assert isinstance(field_info, APIModelFieldInfo)
        return field_info

    def __init__(
        self,
        when: set[Literal['create', 'update', 'store', 'view']] | None = None,
        encrypt: bool = False,
        **kwargs: Any
    ):
        super().__init__(**kwargs)
        self.encrypt = encrypt
        self.when = when or set()

    def is_included(self, mode: str) -> bool:
        return bool(self.when) and mode in self.when


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any


RESPONSE_CONTENT_TYPE: dict[str, Any] = {
    'Content-Type': {
        'schema': {'type': 'string'},
        'required': True,
        'description': (
            "The `Content-Type` representation header is used to "
            "indicate the original media type of the resource "
            "(prior to any content encoding applied for sending)."
        )
    }
}

RESPONSE_ETAG_RESOURCE: dict[str, Any] = {
    'ETag': {
        'schema': {'type': 'string'},
        'required': True,
        'description': (
            "The `ETag` header indicates the returned version of the resource. "
            "It matches the `.metadata.resourceVersion` of the response.\n\n"
            "The `ETag` (or entity tag) HTTP response header is an identifier "
            "for a specific version of a resource. It lets caches be more "
            "efficient and save bandwidth, as the server does not need to "
            "resend a full response if the content was not changed. "
            "Additionally, etags help to prevent simultaneous updates of "
            "a resource from overwriting each other (\"mid-air collisions\")."
        )
    }
}


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import asyncio

from canonical.ext.google.params.credential import get_credential
from canonical.ext.httpx import AsyncClient
from canonical.ext.jose import JOSE
from canonical.ext.google import ServiceAccountTokenSigner
from canonical.ext.google import ServiceAccountTokenVerifier



async def main():
    signer = ServiceAccountTokenSigner(credential=await get_credential())
    async with AsyncClient() as client:
        verifier = ServiceAccountTokenVerifier(http=client)
        obj = await signer.sign_claims(foo=1, bar=2, baz=3)
        jws = JOSE.model_validate(obj)
        print("The token is valid:", await jws.verify(verifier))

if __name__ == '__main__':
    asyncio.run(main())


# Copyright (C) 2023-2024 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from __future__ import annotations
import inspect
import operator
from functools import reduce
from typing import Any
from typing import AsyncContextManager
from typing import AsyncIterable
from typing import cast
from typing import get_args
from typing import overload
from typing import Generator
from typing import Iterable
from typing import Literal
from typing import TypeVar
from typing import TYPE_CHECKING

import aiofiles
import pydantic
import yaml

from canonical.exceptions import DoesNotExist
from canonical.exceptions import Duplicate
from canonical.exceptions import Stale
from canonical.ext.crypto import NullDataEncryption
from canonical.ext.crypto.bases import BaseDEK
from canonical.ext.api.protocols import IObjectIdentifier
from ..annotations import APIResourceType
from ..annotations import Reference
from .resource import BaseResource
from .resourcedto import BaseResourceDTO
if TYPE_CHECKING:
    from ..apimodelinspector import APIModelInspector
    from ..objectmeta import ObjectMeta


B = TypeVar('B')
T = TypeVar('T', bound=APIResourceType)


class BaseResourceRepository:
    inspector: APIModelInspector
    DoesNotExist = DoesNotExist
    Duplicate = Duplicate
    Stale = Stale

    class Cursor(AsyncIterable[B]):
        def __await__(self) -> Generator[Any, Any, list[B]]:
            ...

    def __init__(
        self,
        inspector: APIModelInspector,
        dek: BaseDEK = NullDataEncryption()
    ):
        self.dek = dek
        self.inspector = inspector

    def all(self, model: type[B]) -> Cursor[B]:
        raise NotImplementedError

    def query(
        self,
        model: type[B],
        filters: Iterable[tuple[str, str, Any]] | None = None,
        sort: Iterable[str] | None = None,
        namespace: str | None = None,
        limit: int | None = None,
        kind: str | None = None,
        page_size: int = 10,
        keys: list[Any] | None = None,
        **_: Any
    ) -> Cursor[B]:
        raise NotImplementedError

    async def allocate(self, obj: type[Any]) -> int:
        raise NotImplementedError

    async def exists(
        self,
        key: Reference | Iterable[tuple[str, str, Any]],
        model: Any
    ) -> bool:
        raise NotImplementedError

    @overload
    async def get(
        self,
        key: IObjectIdentifier[T],
        *,
        require: Literal[True]
    ) -> T:
        ...

    @overload
    async def get(
        self,
        key: IObjectIdentifier[T],
        *,
        require: Literal[False]
    ) -> T:
        ...

    @overload
    async def get(
        self,
        key: IObjectIdentifier[T],
        model: None = None,
        *,
        require: Literal[True]
    ) -> T:
        ...

    @overload
    async def get(
        self,
        key: IObjectIdentifier[T],
        model: None = None,
        *,
        require: Literal[False] = False
    ) -> T:
        ...

    @overload
    async def get(
        self,
        key: str,
        model: type[T],
        *,
        require: Literal[True] = True
    ) -> T:
        ...

    @overload
    async def get(
        self,
        key: Reference,
        model: type[T],
        *,
        require: Literal[True]
    ) -> T:
        ...

    @overload
    async def get(
        self,
        key: Reference,
        model: type[T] | None,
        *,
        require: Literal[True] = True
    ) -> T:
        ...

    @overload
    async def get(
        self,
        key: Reference | str,
        model: type[T] | None,
        *,
        require: Literal[False] = False
    ) -> T | None:
        ...

    async def get(
        self,
        key: IObjectIdentifier[T] | Reference | str,
        model: type[T] | None = None,
        *,
        require: bool = True,
        namespace: str | None = None
    ) -> T | None:
        if not isinstance(key, str):
            if model is not None and key.get_model() != model:
                raise TypeError(
                    f"The model referenced by {type(key).__name__} must equal "
                    f"{model.__name__}."
                )
            model = cast(type[T] | None, key.get_model())
        if model is None:
            raise TypeError(
                f"Model should be attached to {type(key).__name__} "
                "or provided with the `model` parameter."
            )
        meta = self.inspector.inspect(model)
            
        # Do some checks if the key is a string. A string key specifies
        # the name of a non-deleted entity. For namespaced resources,
        # this makes the namespace parameter mandatory.
        if isinstance(key, str):
            # Handled in the reference() method.
            key = cast(IObjectIdentifier[T], meta.reference(name=key, namespace=namespace))

        # If the key does not have a model at this point, we assume
        # that it was provided through the model parameter.
        if not key.has_model():
            key.with_model(model)

        # Validate the namespace - must be None for cluster resources,
        # not none and not empty for namespaced resources.
        meta.validate_namespace(key.get_namespace())

        obj = await self.reference(key, model)
        if obj is None and require:
            raise self.DoesNotExist
        if obj is not None:
            await obj.decrypt(self.dek)
        return cast(T | None, obj)

    async def import_documents(
        self,
        kind: Any,
        path: str,
        iterator: bool = True
    ) -> AsyncIterable[APIResourceType]:
        loader: pydantic.TypeAdapter[BaseResourceDTO[APIResourceType]]
        models: list[type[pydantic.BaseModel]] = []
        for model in get_args(kind):
            if not inspect.isclass(model) or not issubclass(model, BaseResource):
                raise TypeError(
                    f"Member '{model.__name__}' of union `kind` is not "
                    "an implementation of BaseResource."
                )
            models.append(model.__create_model__)
        loader = pydantic.TypeAdapter(reduce(operator.or_, models))
        async with aiofiles.open(path) as f: # type: ignore
            documents = cast(list[dict[str, Any]], yaml.safe_load_all(await f.read())) # type: ignore
        for doc in documents:
            new = loader.validate_python(doc)
            old = await self.get(new.key, new.model, require=False)
            if old is not None:
                new = old.replace(new)
            else:
                new = new.factory(uid=await self.allocate(new.model))
            await self.persist(new.model, new, old=old)
            yield new

    async def persist(
        self,
        model: type[pydantic.BaseModel],
        new: T,
        *,
        transaction: Any = None,
        old: T | None = None
    ) -> T:
        async with self.transaction(transaction) as tx:
            assert new.metadata.is_attached()
            if old is None:
                try:
                    old = await self.get(new.key)
                except self.DoesNotExist:
                    old = None
            new = await self.persist_model(new, old=old, transaction=tx)
        return new

    async def persist_metadata(
        self,
        metadata: ObjectMeta[Any],
        transaction: Any | None = None
    ) -> None:
        dao = metadata.dao()
        await self.put(dao.key, dao, transaction=transaction)

    async def persist_model(
        self,
        new: T,
        old: T | None = None,
        transaction: Any = None
    ) -> T:
        if old and old.metadata.generation != new.metadata.generation:
            raise Stale
        updated, changed = new.update(old) # type: ignore
        if not changed:
            return new
        await updated.encrypt(self.dek)
        await self.put(updated.key, updated, transaction=transaction)
        await self.persist_metadata(updated.metadata, transaction=transaction)
        return updated

    async def put(
        self,
        key: IObjectIdentifier[T] | Reference,
        dao: pydantic.BaseModel,
        transaction: Any | None = None
    ) -> None:
        raise NotImplementedError(key)

    async def put_default(
        self,
        key: Reference,
        dao: pydantic.BaseModel,
        transaction: Any | None = None
    ) -> None:
        raise NotImplementedError(key)

    async def reference(
        self,
        key: IObjectIdentifier[T] | Reference,
        model: type[T],
    ) -> T | None:
        raise NotImplementedError

    async def restore(self, instance: T) -> T:
        return instance

    async def on_persisted(self, new: T, old: T | None, transaction: Any | None = None):
        pass

    def transaction(self, transaction: Any | None = None) -> AsyncContextManager[Any]:
        raise NotImplementedError


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import logging

import fastapi

from canonical.ext.iam.protocols import IResourceAuthorizationContext


__all__: list[str] = ['AuthorizationService']

logger: logging.Logger = logging.getLogger('uvicorn')


def AuthorizationService(impl: type[IResourceAuthorizationContext]):
    logger.info("Using %s as authorization implementation.", impl.__name__)

    def f(request: fastapi.Request, ctx: impl = fastapi.Depends(impl)):
        setattr(request.state, 'resource_context', ctx)
        return ctx
    return fastapi.Depends(f)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Iterable

import fastapi
import fastapi.params
from fastapi.exceptions import HTTPException

from canonical.utils.http import MediaTypeSelector


def AcceptedContentType(available: Iterable[str]) -> fastapi.params.Depends:
    selector = MediaTypeSelector(available)

    def f(
        header: str = fastapi.Header(
            default=...,
            alias='Content-Type',
            description=(
                "The `Content-Type` representation header is used to indicate "
                "the original media type of the resource (prior to any "
                "content encoding applied for sending)."
            ),
            examples=list(available)
        )
    ):
        selected = selector.select(header)
        if selected is None:
            raise HTTPException(
                status_code=415,
                detail=(
                    "The server refuses to accept the request because "
                    "the payload format is in an unsupported format."
                )
            )
        return selected

    return fastapi.Depends(f)


# Copyright (C) 2023-2024 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import ClassVar
from typing import Generic
from typing import TypeVar
from typing import TYPE_CHECKING

from canonical.protocols import ITyped
if TYPE_CHECKING:
    from .rootresource import ResourceType


__all__: list[str] = [
    'PrimaryKey'
]

M = TypeVar('M')


class PrimaryKey(ITyped[M], Generic[M]):
    model: type[M]
    api_group: ClassVar[str]
    plural: ClassVar[str]
    name: str
    namespace: str | None

    @classmethod
    def new(cls, t: type[M], **kwargs: Any) -> ITyped[M]:
        return cls.typed(t)(**kwargs) # type: ignore

    @classmethod
    def typed(cls, t: type['ResourceType']) -> type[ITyped[M]]:
        return type(f'{t.__name__}PrimaryKey', (cls,), {
            'api_group': t.group,
            'model': t,
            'plural': t.plural,
        })

    def __init__(
        self,
        name: str,
        namespace: str | None = None,
        server: str | None = None
    ):
        self.name = name
        self.namespace = namespace
        self.server = server

    def get_namespace(self):
        return self.namespace

    def get_type(self) -> type[M]:
        return self.model

    def __repr__(self):
        return f"<Key('{self.model.__name__}', {repr(self.name)}, namespace={repr(self.namespace)})"


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

from pydantic_core import CoreSchema
from pydantic_core import core_schema
from pydantic.json_schema import JsonSchemaValue
from pydantic import GetJsonSchemaHandler

from .basepermission import BasePermission


__all__: list[str] = [
    'Permission'
]


class Permission(BasePermission):

    @classmethod
    def __get_pydantic_core_schema__(cls, *_: Any) -> CoreSchema:
        return core_schema.json_or_python_schema(
            json_schema=core_schema.int_schema(),
            python_schema=core_schema.union_schema([
                core_schema.is_instance_schema(cls),
                core_schema.chain_schema([
                    core_schema.str_schema(max_length=64),
                    core_schema.no_info_plain_validator_function(cls)
                ])
            ]),
            serialization=core_schema.plain_serializer_function_ser_schema(str),
        )

    @classmethod
    def __get_pydantic_json_schema__(
        cls,
        _: CoreSchema,
        handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        return handler(core_schema.str_schema())

    def matches(self, permission: str) -> bool:
        return str(self) == str(permission)

    def __repr__(self) -> str: # pragma: no cover
        return f'{type(self).__name__}({str(self)})'


# Copyright (C) 2023-2024 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from __future__ import annotations
import datetime
import inspect
import logging
from typing import cast
from typing import Any
from typing import ClassVar
from typing import Generic
from typing import Literal
from typing import Self
from typing import TypeVar
from typing import TYPE_CHECKING

import pydantic
import yaml

from canonical.exceptions import Immutable
from canonical.exceptions import Inconsistent
from canonical.exceptions import Stale
from canonical.ext.api import ObjectReference
from canonical.protocols import IStorage
from canonical.protocols import ITransaction
from canonical.utils import deephash
from .apiresourcelist import APIResourceList
from .listbase import ListBase
from .namespacedobjectmeta import NamespacedObjectMeta
from .objectmeta import ObjectMeta
from .resourcekey import ResourceKey
from .resourcemeta import ResourceMeta
from .transientmeta import TransientMeta
if TYPE_CHECKING:
    from .baseresourcerepository import BaseResourceRepository
    from .rootresource import RootResource

__all__: list[str] = [
    'Resource'
]

M = ObjectMetaType = TypeVar('M', bound=ObjectMeta[Any] | TransientMeta)
S = TypeVar('S', bound='Resource[Any]')
V = TypeVar('V')
NOT_PROVIDED: object = object()


class Resource(pydantic.BaseModel, Generic[M]):
    _storage: BaseResourceRepository | None = pydantic.PrivateAttr(default=None)
    List: ClassVar[type[ListBase[Any, Any]]]
    metadata_class: ClassVar[type[ObjectMeta[Any] | TransientMeta]]
    model_config = {'populate_by_name': True}
    base_path: ClassVar[str]
    group: ClassVar[str]
    logger: ClassVar[logging.Logger] = logging.getLogger('canonical.ext.resource')
    plural: ClassVar[str]
    InputModel: ClassVar[type[Self]]
    __meta__: ClassVar[ResourceMeta]

    api_version: str = pydantic.Field(
        default=...,
        alias='apiVersion',
        title="API Version",
        description=(
            "The `apiVersion` field defines the versioned schema of this "
            "representation of an object. Servers should convert recognized "
            "schemas to the latest internal value, and may reject "
            "unrecognized values."
        ),
        frozen=True
    )

    kind: str = pydantic.Field(
        default=...,
        description=(
            "Kind is a string value representing the REST resource this "
            "object represents. Servers may infer this from the endpoint "
            "the client submits requests to. Cannot be updated. In `CamelCase`."
        ),
        frozen=True
    )

    metadata: M = pydantic.Field(
        default=...,
        title='Metadata',
        description=(
            "`ObjectMeta` is metadata that all persisted resources "
            "must have, which includes all objects users must create."
        )
    )

    @property
    def scoped(self):
        n = self.metadata.name
        if self.is_namespaced():
            assert isinstance(self.metadata, NamespacedObjectMeta)
            n = f'{self.metadata.namespace}/{self.metadata.name}'
        return n

    @classmethod
    def is_namespaced(cls) -> bool:
        metadata_cls = cls.model_fields['metadata'].annotation
        if not inspect.isclass(metadata_cls):
             return False
        assert metadata_cls in {ObjectMeta, NamespacedObjectMeta, TransientMeta}\
            or issubclass(metadata_cls, (ObjectMeta, NamespacedObjectMeta, TransientMeta))
        return metadata_cls.is_namespaced() # type: ignore

    @classmethod
    def is_destroyable(cls) -> bool:
        """Return a boolean indicating if the resource may be destroyed
        by a client.
        """
        return True

    @classmethod
    def is_purgable(cls) -> bool:
        """Return a boolean indicating if the resources may be purged
        by a client.
        """
        return True

    @classmethod
    def new(cls, name: Any, **params: Any):
        return cls.model_validate({
            **params,
            'metadata': {
                'name': name
            }
        })

    @property
    def key(self):
        assert not isinstance(self.metadata, TransientMeta)
        ref = ObjectReference(
            api_version=self.api_version,
            kind=self.kind,
            name=self.metadata.name,
            namespace=self.metadata.get_namespace() or '',
            resource_version=self.metadata.resource_version or '',
            uid=self.metadata.uid
        )
        return ref.attach(self.metadata).with_model(type(self))

    @property
    def relname(self) -> str:
        return f'{self.plural}/{self.metadata.name}'

    @classmethod
    def add_to_root(cls, root: type[RootResource[Any]], meta: ResourceMeta) -> None:
        assert cls.model_fields['metadata'].annotation
        cls.model_fields['api_version'].default = meta.api_version
        cls.model_fields['api_version'].annotation = Literal[f'{meta.api_version}']
        cls.model_fields['kind'].default = root.__name__
        cls.model_fields['kind'].annotation = Literal[f'{root.__name__}']
        assert cls.model_rebuild(force=True)
        if not hasattr(cls, 'InputModel'):
            # TODO: This will cause weird results if a subclass of Resource
            # is added to multiple RootResource subclasses.
            cls.InputModel = ResourceMeta.build_input_model(
                model=cls,
                meta=cls.model_fields['metadata'].annotation,
                fields=cls.model_fields
            )

    @classmethod
    def has_state(cls) -> bool:
        return False

    @classmethod
    def model_validate_input(cls, data: dict[str, Any]) -> Self:
        return cls.model_validate(data)

    @classmethod
    def setup_meta(cls, version: str, contribute: bool = True):
        meta = ResourceMeta.fromqualname(cls, version)
        if contribute:
            meta.contribute_to_class(cls, cls.model_fields)
        return meta

    def __init_subclass__(cls, **kwargs: Any):
        super().__init_subclass__()

    @classmethod
    def __pydantic_init_subclass__(cls, **kwargs: Any):
        super().__init_subclass__()
        qualname = kwargs.get('version')
        if qualname:
            cls.setup_meta(qualname)

            # Create the list type.
            cls.List = type(f'{cls.__name__}List', (ListBase[Literal[f'{cls.__name__}List'], cls],), { # type: ignore
                'items': pydantic.Field(
                    default_factory=list,
                    description=(
                        "The `items` member contains an array "
                        f"of `{cls.__name__}` objects."
                    )
                ),
                '__annotations__': {'items': list[cls]}
            })
            cls.List.model_fields['kind'].default = f'{cls.__name__}List'
            cls.List.model_rebuild()

            metadata_class = cls.model_fields['metadata'].annotation
            if not inspect.isclass(metadata_class):
                return None
            assert metadata_class in {ObjectMeta, NamespacedObjectMeta, TransientMeta}\
                or issubclass(metadata_class, (ObjectMeta, NamespacedObjectMeta, TransientMeta))
            cls.metadata_class = metadata_class
            cls.model_fields['metadata'].default_factory = metadata_class.default # type: ignore
            #cls.model_rebuild()

    def can_change(self, old: Self) -> bool:
        return True

    def get_comparison_fields(self) -> set[str]:
        return {'spec'}

    def get_mutable_data(self) -> dict[str, Any]:
        return self.model_dump(
            mode='json',
            include=self.get_mutable_fields()
        )

    def get_mutable_fields(self) -> set[str]:
        return set()

    def get_namespace(self) -> str | None:
        return self.metadata.get_namespace()

    def in_namespace(self, namespace: str | None) -> bool:
        return self.metadata.in_namespace(namespace)

    def is_changed(self, old: Self) -> bool:
        a = deephash(self.model_dump(mode='json', include=self.get_comparison_fields()))
        b = deephash(old.model_dump(mode='json', include=old.get_comparison_fields()))
        return a != b

    def is_created(self):
        return bool(self.metadata.resource_version)

    def is_consistent(self) -> bool:
        return True

    def is_persistable(self) -> bool:
        return isinstance(self.metadata, (ObjectMeta, NamespacedObjectMeta))

    def model_post_init(self, _: Any) -> None:
        self.metadata.attach(self)

    def model_dump_yaml(self, indent: int =2, **kwargs: Any) -> str:
        return yaml.safe_dump(  # type: ignore
            self.model_dump(mode='json', by_alias=True, **kwargs),
            default_flow_style=False
        )

    async def persist(
        self,
        storage: Any | None = None,
        mode: Literal['create', 'replace'] = 'create',
        transaction: ITransaction | None = None
    ) -> Self:
        return await self.persist_key(self.key, storage, mode, transaction)

    async def persist_key(
        self,
        key: ResourceKey,
        storage: BaseResourceRepository | None = None,
        mode: Literal['create', 'replace', 'update'] = 'create',
        transaction: ITransaction | None = None,
        transfer: bool = False
    ) -> Self:
        assert isinstance(self.metadata, (ObjectMeta, NamespacedObjectMeta))
        if not self.is_consistent():
            raise Inconsistent(
                f"The {type(self).__name__} is in an inconsistent state."
            )
        storage = storage or self._storage
        if storage is None:
            raise TypeError(
                f'{type(self).__name__} is not attached to any storage '
                'and the `storage` parameter is None.'
            )
        try:
            old = await storage.get(key, type(self))
            old.metadata.updated = datetime.datetime.now(datetime.timezone.utc)
            assert old is not None
        except storage.DoesNotExist:
            if mode != 'create':
                raise
            old = None

            # We are creating a new object here, so the generation
            # should be bumped to the first version.
            self.metadata.generation = 1

            # Clear any existing values for updated, uid
            self.metadata.updated = self.metadata.created
            self.metadata.uid = await storage.allocate(key.get_model())

            # The deleted field can not be set on new objects,
            # because there is nothing to delete.
            if self.metadata.deleted: # type: ignore
                raise Immutable(
                    "The .metadata.deleted field must not be set on new "
                    f"{type(self).__name__} objects."
                )
        if mode == 'create' and old is not None:
            raise storage.Duplicate
        if old is not None and mode == 'update':
            nm = cast(ObjectMeta[Any], self.metadata)
            om = cast(ObjectMeta[Any], old.metadata)
            assert isinstance(old.metadata, (ObjectMeta, NamespacedObjectMeta))

            # If the generation is non-zero, then the generation of the persisted
            # object MUST match it.
            if nm.generation and (nm.generation != om.generation):
                raise Stale

            # Same applies to the resource version.
            if nm.resource_version and (nm.resource_version != om.resource_version):
                raise Stale

            # This should never happen, but its fine to check anyway.
            if (getattr(nm, 'namespace', None) != getattr(om, 'namespace', None))\
            and not transfer:
                raise Immutable(
                    "The namespace of an object can not be changed."
                )

            # If the existing state does not allow transition to the
            # new state, throw an immutable exception.
            if not self.can_change(old):
                raise Immutable(
                    f"{type(self).__name__} does not allow the specified "
                    "changes."
                )

            # Merge the metadata of the new instance into the old instance.
            self.metadata = om.merge(nm) # type: ignore

            # Bump the generation (TODO: only bump if there are actual)
            # changes.
            if self.is_changed(old):
                self.metadata.generation += 1

        if old is not None and mode == 'replace':
            assert self.replacable()
            self.metadata.update(self, cast(Any, old.metadata), mode=mode)
            # TODO: on stateful resources, clear the status.
            self._on_replaced(old)

        # Resource version must be updated on every write.
        self.metadata.update_resource_version(self)
        return await storage.persist(self, transaction=transaction, model=key.get_model())

    def replacable(self) -> bool:
        return True

    def _on_replaced(self, old: Self):
        pass


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import dataclasses
import datetime
from typing import Any
from typing import Literal

from .base import BaseCache


class MemoryCache(BaseCache):
    __module__: str = 'canonical.ext.cache'
    _global: dict[str, 'CachedObject'] = {}
    objects: dict[str, 'CachedObject']

    @dataclasses.dataclass
    class CachedObject:
        expires: datetime.datetime | None
        value: bytes

        def is_expired(self) -> bool:
            now = datetime.datetime.now(datetime.timezone.utc)
            return self.expires is not None and (self.expires < now)

    def __init__(self, scope: Literal['global', 'local'], *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.scope = scope
        self.objects = {}
        if scope == 'global':
            self.objects = MemoryCache._global

    async def fetch(self, key: str, *args: Any, **kwargs: Any) -> bytes | None:
        value = None
        obj = self.objects.get(key)
        if obj is not None and not obj.is_expired():
            value = obj.value
        return value

    async def put(
        self,
        key: str,
        value: bytes,
        *_: Any,
        ttl: int | None = None,
        **__: Any
    ) -> None:
        now = datetime.datetime.now(datetime.timezone.utc)
        expires = None
        if ttl:
            expires = now + datetime.timedelta(seconds=ttl)
        self.objects[key] = self.CachedObject(expires, value)


# Copyright (C) 2023-2024 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import ClassVar
from typing import Self

from .apimodel import APIModel
from .apiversionedmeta import APIVersionedMeta
from .fields import APIVersion
from .fields import Kind
from .registry import register


__all__: list[str] = [
    'APIVersioned'
]


class APIVersioned(APIModel, abstract=True):
    __abstract__ = True
    __descriptor_class__: ClassVar[type[APIVersionedMeta[Self]]] = APIVersionedMeta
    __meta__: ClassVar[Any]
    api_version: APIVersion
    kind: Kind

    @classmethod
    def __pydantic_init_subclass__(
        cls,
        namespaced: bool = False,
        descriptor: type[APIVersionedMeta[Self]] | None = None,
        api_version: str | None = None,
        group: str | None = None,
        version: str | None = None,
        kind: str | None = None,
        plural: str | None = None,
        short: str | None = None,
        **kwargs: Any
    ):
        meta: APIVersionedMeta[Any] | None = getattr(cls, '__meta__', None)
        if descriptor is not None:
            cls.__descriptor_class__ = descriptor
        if meta:
            kwargs.update({
                'group': meta.api_group,
                'plural': meta.plural,
                'version': meta.version
            })
        super().__pydantic_init_subclass__(namespaced=namespaced, **kwargs)

    def __init_subclass__(
        cls,
        group: str | None = None,
        version: str | None = None,
        kind: str | None = None,
        plural: str | None = None,
        short: str | None = None,
        namespaced: bool = False,
        **kwargs: Any
    ):
        if (group is not None and version is not None)\
        and not kwargs.get('abstract'):
            meta = cls.__descriptor_class__( # type: ignore
                group=group,
                version=version,
                kind=kind or cls.__name__,
                plural=plural,
                short=short,
                namespaced=namespaced,
            )
            meta.contribute_to_class(cls, cls.model_fields)
            assert hasattr(cls, '__meta__')
            register(cls.__meta__)
        return super().__init_subclass__(namespaced=namespaced, **kwargs)

    def has_state(self) -> bool:
        raise NotImplementedError


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import cast

from canonical.ext.api import APIResourceType
from canonical.ext.fastapi.params import ResourceRepository
from canonical.ext.oauth.protocols import IRegisteredClient
from canonical.ext.oauth.types import URLSafeClientID
from .baseendpoint import BaseEndpoint


class AuthorizationServerEndpoint(BaseEndpoint):
    client_cache_ttl: int = 3600
    resources: ResourceRepository
    _client: IRegisteredClient | None

    def get_client_meta(self) -> tuple[URLSafeClientID, str, type]:
        raise NotImplementedError

    async def get_client(self) -> IRegisteredClient | None:
        client_id, cache_key, model = self.get_client_meta()
        client = cast(APIResourceType | None, await self.cache.get(cache_key, model))
        if client is None:
            client = await self.resources.get(client_id.ref)
            await self.cache.set(
                key=cache_key,
                value=client,
                encoder=model,
                ttl=self.client_cache_ttl
            )
        self.client_resource = client
        return cast(IRegisteredClient, getattr(client, 'spec'))


    async def setup(self):
        self._client = await self.get_client()
        setattr(self.request.state, 'oauth_client', self._client)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import datetime
import inspect
from typing import Any
from typing import ClassVar
from typing import Generic
from typing import TypeVar
from typing import TYPE_CHECKING

import pydantic

from canonical.exceptions import ProgrammingError
if TYPE_CHECKING:
    from .resourcestatus import ResourceStatus


R = TypeVar('R')
S = TypeVar('S', bound='ResourceStatus[Any]')
RESERVED_FIELDS: set[str] = {'uid', 'namespace', 'name', 'parent'}


class ConditionType(pydantic.BaseModel, Generic[S]):
    """Describes the condition of a :class:`VersionedResource`."""
    model_config = {'populate_by_name': True}
    StorageModel: ClassVar[type[pydantic.BaseModel]]
    __handler_args__: ClassVar[set[str]]
    __storage__: ClassVar[bool] = False
    _can_replay: ClassVar[bool] = False
    _uid: int | None = pydantic.PrivateAttr(default=None)

    timestamp: datetime.datetime = pydantic.Field(
        default=...,
        description=(
            "Specifies the date and time at which the condition "
            "emerged."
        )
    )

    message: str = pydantic.Field(
        default='',
        description=(
            "A human readable message indicating details about "
            "the transition. This may be an empty string."
        )
    )

    observed_generation: int = pydantic.Field(
        default=...,
        alias='observedGeneration',
        description=(
            "Represents the `.metadata.generation` that the "
            "condition was set based upon. For instance, if "
            "`.metadata.generation` is currently 12, but "
            "the `.status.conditions[x].observedGeneration` "
            "is 9, the condition is out of date with respect "
            "to the current state of the instance."
        )
    )

    status: str = pydantic.Field(
        default=...,
        description=(
            "Contains a programmatic identifier indicating the "
            "status of the condition's last transition. "
            "Producers of specific condition types may define "
            "expected values and meanings for this field, "
            "and whether the values are considered a guaranteed "
            "API. This field may not be empty."
        )
    )

    @property
    def uid(self):
        return self._uid

    def __init_subclass__(cls, **kwargs: Any):
        return super().__init_subclass__()

    @classmethod
    def __pydantic_init_subclass__(cls, status: str | None = None, **kwargs: Any) -> None:
        reserved = set(cls.model_fields) & RESERVED_FIELDS
        if reserved and not cls.__storage__:
            raise ProgrammingError(
                f"{cls.__name__} contains reserved fields: "
                f"{sorted(reserved)}."
            )

        # Determine which arguments the handler accepts.
        if inspect.ismethod(cls.handle):
            raise ProgrammingError(
                f'Subclasses must override {cls.__name__}.handle().'
            )

        sig = inspect.signature(cls.handle)
        if 'replay' in sig.parameters:
            if sig.parameters['replay'].annotation != bool:
                raise ProgrammingError(
                    "The `replay` argument must be annotated as a boolean "
                    "(replay: bool). This parameter is passed in via the "
                    "base class and indicates if a condition is being "
                    "replayed."
                )
            cls._can_replay = True

        # TODO: Do not hardcode the identifier type.
        if not cls.__storage__:
            cls.StorageModel = type(
                cls.__name__,
                (cls,),
                {
                    '__annotations__': {
                        'name': str,
                        'namespace': str | None,
                        'parent': int,
                        'uid': int,
                    },
                    '__storage__': True
                }
            )

    @classmethod
    def can_replay(cls) -> bool:
        return cls._can_replay

    def handle(self, state: S, **kwargs: Any) -> None:
        raise NotImplementedError

    def apply(self, state: S, replay: bool = False) -> None:
        kwargs = self.model_dump(include=self.__handler_args__)
        if self.can_replay():
            kwargs['replay'] = replay
        return self.handle(state, **kwargs)

    def can_transition(self, state: S):
        return True

    def is_dirty(self):
        return self._uid is None

    def is_final(self) -> bool:
        return False


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import httpx
import pydantic

from canonical.ext.api import ResourceSpec
from ..models import OIDCProvider
from ..models import ServerMetadata


class BaseProviderSpec(ResourceSpec):
    metadata: ServerMetadata | OIDCProvider = pydantic.Field(
        default=...,
        alias='serverMetadata',
        description=(
            "The metadata describing the authorization server."
        )
    )

    def is_discoverable(self):
        return isinstance(self.metadata, OIDCProvider)

    async def discover(self, client: httpx.AsyncClient):
        if not isinstance(self.metadata, ServerMetadata):
            self.metadata = await self.metadata.discover(client)
        return self.metadata


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Literal

import pydantic

from .grantmodel import GrantModel


class JWTBearerRequest(GrantModel[Literal['urn:ietf:params:oauth:grant-type:jwt-bearer']]):
    assertion: str = pydantic.Field(
        default=...,
        title="Assertion (JWT)",
        description=(
            "A compact encoded JSON Web Token (JWT).\n\n"
            "The value of the `assertion` parameter **must** "
            "contain a single JWT."
        )
    )

    scope: str | None = pydantic.Field(
        default=None,
        title="Scope",
        description=(
            "The scope of the access request."
        )
    )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import enum


class KeyOpEnum(str, enum.Enum):
    bits    = 'deriveBits'
    decrypt = 'decrypt'
    derive  = 'deriveBits'
    encrypt = 'encrypt'
    sign    = 'sign'
    verify  = 'verify'
    unwrap  = 'unwrapKey'
    wrap    = 'wrapKey'


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import fastapi

from .credential import Credential as GoogleCredential
from .datastore import Datastore
from .datastoreclient import DatastoreClient


__all__: list[str] = [
    'Datastore',
    'DatastoreClient',
    'GoogleCredential',
    'GoogleDatastore',
]


def GoogleDatastore(
    namespace: str | None = None
):
    def f(request: fastapi.Request):
        setattr(request.state, 'google_datastore_namespace', namespace)

    return fastapi.Depends(f)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Generic
from typing import TypeVar

import pydantic

from canonical.ext.api import ObjectMeta
from canonical.ext.api import ResourceSpec
from canonical.lib.types import HTTPResourceLocator
from canonical.ext.oauth.models.requests import AuthorizationRequest
from canonical.ext.oauth.models.responses import TokenResponse
from canonical.ext.oauth.types import RedirectURI


M = TypeVar('M')


class AuthorizationStateSpec(ResourceSpec, Generic[M]):
    """Describes the state of an OAuth 2.x/OpenID Connect
    authorization request.
    """
    issuer: HTTPResourceLocator = pydantic.Field(
        default=...,
        title="Issuer",
        description=(
            "The issuer identifier of the authorization server."
        )
    )

    authorization_endpoint: HTTPResourceLocator = pydantic.Field(
        default=...,
        title="Authorization Endpoint",
        alias='authorizationEndpoint',
        description=(
            "The authorization endpoint to which the authorization "
            "request is issued, without query parameters."
        )
    )

    token_endpoint: HTTPResourceLocator = pydantic.Field(
        default=...,
        title="Token Endpoint",
        alias='tokenEndpoint',
        description=(
            "The token endpoint from which the token is obtained after "
            "receiving an authorization code."
        )
    )

    client_id: str = pydantic.Field(
        default=...,
        alias='clientId',
        title="Client ID",
        description=(
            "The client identifier of the **local** client. If "
            "`.clientId` equals `.params.client_id`, then this "
            "server is obtaining the access token. Otherwise, it "
            "is acting as a proxy or intermediate on the behalf "
            "of another server. In that case, the `.clientId` "
            "parameter points to the local client configuration "
            "that the server maintains."
        )
    )

    params: AuthorizationRequest = pydantic.Field(
        default=...,
        title="Parameters",
        description=(
            "The parameters of the authorization request, as specified "
            "by the OAuth 2.x/OpenID Connect specifications. Note that "
            "the fields in this object are not `pascalCase` to maintain "
            "conformance with the specifications."
        )
    )

    redirect_mode: M = pydantic.Field(
        default='proxy',
        alias='redirectMode',
        title="Redirect Mode",
        description=(
            "Specifies the redirection mode.\n\nIf `.mode` is `proxy`, then "
            "the obtained authorization code and other parameters are "
            "forwared unmodified to the clients' redirection endpoint (as "
            "specified by the `.redirectUri` parameter.\n\n"
            "With the `issue` mode, this server obtains the authorization "
            "code, and issues its own authorization code to pass back "
            "to the client.\n\nThe `obtain` mode indicates that the server "
            "itself wants to obtain the token (i.e. it is acting as a client "
            "application)."
        )
    )

    redirect_uri: RedirectURI | None = pydantic.Field(
        default=None,
        title="Redirect URI",
        alias='redirectUri',
        description=(
            "The redirection endpoint, if it differs from the `redirect_uri` "
            "query parameter in the authorization request.\n\n"
            "If the `redirect_uri` parameter was not included in the authorization "
            "request, this field holds the value that client selected."
        )
    )

    def get_user_agent_redirect_uri(
        self,
        metadata: ObjectMeta[str]
    ) -> str:
        raise NotImplementedError

    def get_redirect_uri(self) -> RedirectURI | None:
        """Return the redirect URI for use with the given
        audience.
        """
        uri = None
        if self.params.redirect_uri:
            uri = self.params.redirect_uri
        return uri

    def get_return_url(
        self,
        code: str,
        iss: str | None = None
    ) -> str:
        raise NotImplementedError
        

    def on_obtained(self, response: TokenResponse) -> dict[str, Any]:
        raise NotImplementedError


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Generic
from typing import TypeVar

import pydantic

from ..bases import BaseBuilder


T = TypeVar('T', bound=pydantic.BaseModel)


class BaseModelClassBuilder(BaseBuilder[T], Generic[T]):
    children: list[type[pydantic.BaseModel]]

    def rebuild(self):
        assert self.model.model_rebuild(force=True)
        self.fields = self.model.model_fields

    def setup(self, model: type[T], **kwargs: Any) -> None: # type: ignore
        self.fields = model.model_fields

    def set_field_annotation(self, model: pydantic.BaseModel, name: str, annotation: Any):
        model.model_fields[name].annotation = annotation


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from __future__ import annotations
from typing import Any
from typing import Generic
from typing import Protocol
from typing import TypeVar

import httpx

from canonical.ext.oauth.types import Error
from canonical.ext.oauth.types import ProtocolViolation


R = TypeVar('R', covariant=True)


class IObtainingClient(Protocol, Generic[R]):

    @property
    def id(self) -> str:
        ...

    def configure(self, **kwargs: Any) -> IObtainingClient[R]:
        """Configure a new instance with the given parameters."""
        ...

    async def obtain(
        self,
        grant: Any,
        http: httpx.AsyncClient | None = None
    ) -> R | Error | ProtocolViolation:
        ...


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic

from .iso3166 import ISO3166Alpha2
from .text import Text


class BaseAddress(pydantic.BaseModel):
    model_config = {'populate_by_name': True}

    address1: Text = pydantic.Field(
        default=...
    )

    address2: Text | None = pydantic.Field(
        default=None
    )

    address3: Text | None = pydantic.Field(
        default=None
    )

    city: Text = pydantic.Field(
        default=...
    )

    postal_code: Text = pydantic.Field(
        default=...,
        alias='postalCode'
    )

    country_code: ISO3166Alpha2 = pydantic.Field(
        default=...,
        alias='countryCode'
    )


# Copyright (C) 2023-2024 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Protocol
from typing import Self

import pydantic


class IResource(Protocol):

    def model_validate(self, data: dict[str, Any]) -> Self: ...
    def replace(self, instance: pydantic.BaseModel) -> Self: ...


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Generic
from typing import TypeVar

import pydantic

from canonical.ext.api import APIResourceModel
from .rolereference import RoleReference
from .subject import Subject


RefType = TypeVar('RefType')


class BaseRoleBinding(APIResourceModel[str], Generic[RefType], abstract=True):
    subjects: list[Subject] = pydantic.Field(
        default=...,
        description=(
            "Holds references to the objects the role applies to."
        ),
        min_length=1
    )

    role_ref: RoleReference[RefType] = pydantic.Field(
        default=...,
        alias='roleRef',
        description=(
            "The `roleRef` property must reference a `Role` in the current namespace "
            "or a `GlobalRole` in the global namespace. If `roleRef` cannot be resolved, "
            "the Authorizer must return an error."
        )
    )

    def is_global(self) -> bool:
        return self.role_ref.is_global()


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import enum


class JWASigningAlgorithm(str, enum.Enum):
    hs256 = 'HS256'
    hs384 = 'HS384'
    hs512 = 'HS512'
    rs256 = 'RS256'
    rs384 = 'RS384'
    rs512 = 'RS512'
    es256 = 'ES256'
    es384 = 'ES384'
    es512 = 'ES512'
    eddsa = 'EdDSA'